"""
ARTIO-specific fields




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.fields.field_info_container import \
    FieldInfoContainer
from yt.units.yt_array import \
    YTArray
from yt.utilities.physical_constants import \
    boltzmann_constant_cgs, \
    amu_cgs
import numpy as np

b_units = "code_magnetic"
ra_units = "code_length / code_time**2"
rho_units = "code_mass / code_length**3"
vel_units = "code_velocity"
# NOTE: ARTIO uses momentum density.
mom_units = "code_mass / (code_length**2 * code_time)"
en_units = "code_mass*code_velocity**2/code_length**3"
p_units = "code_mass / (code_length * code_time**2)"

class ARTIOFieldInfo(FieldInfoContainer):
    known_other_fields = (
        ("HVAR_GAS_DENSITY", (rho_units, ["density"], None)),
        ("HVAR_GAS_ENERGY", (en_units, ["total_energy"], None)),
        ("HVAR_INTERNAL_ENERGY", (en_units, ["thermal_energy"], None)),
        ("HVAR_PRESSURE", (p_units, ["pressure"], None)),
        ("HVAR_MOMENTUM_X", (mom_units, ["momentum_x"], None)),
        ("HVAR_MOMENTUM_Y", (mom_units, ["momentum_y"], None)),
        ("HVAR_MOMENTUM_Z", (mom_units, ["momentum_z"], None)),
        ("HVAR_GAMMA", ("", ["gamma"], None)),
        ("HVAR_METAL_DENSITY_Ia", (rho_units, ["metal_ia_density"], None)),
        ("HVAR_METAL_DENSITY_II", (rho_units, ["metal_ii_density"], None)),
        ("VAR_POTENTIAL", ("", ["potential"], None)),
        ("VAR_POTENTIAL_HYDRO", ("", ["gas_potential"], None)),
        ("RT_HVAR_HI", (rho_units, ["HI density"], None)),
        ("RT_HVAR_HII", (rho_units, ["HII density"], None)),
        ("RT_HVAR_HeI", (rho_units, ["HeI density"], None)),
        ("RT_HVAR_HeII", (rho_units, ["HeII density"], None)),
        ("RT_HVAR_HeIII", (rho_units, ["HeIII density"], None)),
    )

    known_particle_fields = (
        ("POSITION_X", ("code_length", ["particle_position_x"], None)),
        ("POSITION_Y", ("code_length", ["particle_position_y"], None)),
        ("POSITION_Z", ("code_length", ["particle_position_z"], None)),
        ("VELOCITY_X", (vel_units, ["particle_velocity_x"], None)),
        ("VELOCITY_Y", (vel_units, ["particle_velocity_y"], None)),
        ("VELOCITY_Z", (vel_units, ["particle_velocity_z"], None)),
        ("MASS", ("code_mass", ["particle_mass"], None)),
        ("PID", ("", ["particle_index"], None)),
        ("SPECIES", ("", ["particle_type"], None)),
        ("BIRTH_TIME", ("", [], None)),  # code-units defined as dimensionless to
                                         # avoid incorrect conversion
        ("INITIAL_MASS", ("code_mass", ["initial_mass"], None)),
        ("METALLICITY_SNIa", ("", ["metallicity_snia"], None)),
        ("METALLICITY_SNII", ("", ["metallicity_snii"], None)),
    )

    def setup_fluid_fields(self):
        unit_system = self.ds.unit_system
        def _get_vel(axis):
            def velocity(field, data):
                return data["momentum_%s" % axis]/data["density"]
            return velocity
        for ax in 'xyz':
            self.add_field(("gas", "velocity_%s" % ax), sampling_type="cell",
                           function = _get_vel(ax),
                           units = unit_system["velocity"])

        def _temperature(field, data):
            tr = data["thermal_energy"]/data["density"]
            # We want this to match *exactly* what ARTIO would compute
            # internally.  We therefore use the exact values that are internal
            # to ARTIO, rather than yt's own internal constants.
            mH  = 1.007825*amu_cgs
            mHe = 4.002602*amu_cgs
            Yp    = 0.24
            XH    = 1.0 - Yp
            XHe   = 0.25*Yp
            mb = XH*mH + XHe*mHe
            wmu   = 4.0/(8.0-5.0*Yp)
            # Note that we have gamma = 5.0/3.0 here
            tr *= (data["gamma"] - 1.0)
            tr *= wmu
            tr *= mb/boltzmann_constant_cgs
            return tr
        # TODO: The conversion factor here needs to be addressed, as previously
        # it was set as:
        # unit_T = unit_v**2.0*mb / constants.k
        self.add_field(("gas", "temperature"), sampling_type="cell",  function = _temperature,
                       units = unit_system["temperature"])

        # Create a metal_density field as sum of existing metal fields.
        flag1 = ("artio", "HVAR_METAL_DENSITY_Ia") in self.field_list
        flag2 = ("artio", "HVAR_METAL_DENSITY_II") in self.field_list
        if flag1 or flag2:
            if flag1 and flag2:
                def _metal_density(field, data):
                    tr = data['metal_ia_density'].copy()
                    np.add(tr, data["metal_ii_density"], out=tr)
                    return tr
            elif flag1 and not flag2:
                def _metal_density(field, data):
                    tr = data["metal_ia_density"]
                    return tr
            else:
                def _metal_density(field, data):
                    tr = data["metal_ii_density"]
                    return tr
            self.add_field(("gas","metal_density"), sampling_type="cell",
                           function=_metal_density,
                           units=unit_system["density"],
                           take_log=True)

    def setup_particle_fields(self, ptype):
        if ptype == "STAR":
            def _creation_time(field,data):
                return YTArray(data.ds._handle.tphys_from_tcode_array(data["STAR","BIRTH_TIME"]),"yr")

            def _age(field, data):
                return data.ds.current_time - data["STAR","creation_time"]

            self.add_field((ptype, "creation_time"), sampling_type="particle",  function=_creation_time, units="yr")
            self.add_field((ptype, "age"), sampling_type="particle",  function=_age, units="yr")

            if self.ds.cosmological_simulation:
                def _creation_redshift(field,data):
                    return 1.0/data.ds._handle.auni_from_tcode_array(data["STAR","BIRTH_TIME"]) - 1.0

                self.add_field((ptype, "creation_redshift"), sampling_type="particle",  function=_creation_redshift)

        super(ARTIOFieldInfo, self).setup_particle_fields(ptype)



"""
The basic field info container resides here.  These classes, code specific and
universal, are the means by which we access fields across YT, both derived and
native.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
from numbers import Number as numeric_type

from yt.extern.six import string_types
from yt.funcs import mylog, only_on_root
from yt.geometry.geometry_handler import \
    is_curvilinear
from yt.units.unit_object import Unit
from yt.units.dimensions import dimensionless
from .derived_field import \
    DerivedField, \
    NullFunc, \
    TranslationFunc
from yt.utilities.exceptions import \
    YTFieldNotFound
from .field_plugin_registry import \
    field_plugins
from .particle_fields import \
    add_union_field, \
    particle_deposition_functions, \
    particle_vector_functions, \
    particle_scalar_functions, \
    standard_particle_fields, \
    add_volume_weighted_smoothed_field, \
    sph_whitelist_fields

def tupleize(inp):
    if isinstance(inp, tuple):
        return inp
    # prepending with a '?' ensures that the sort order is the same in py2 and
    # py3, since names of field types shouldn't begin with punctuation
    return ('?', inp, )

class FieldInfoContainer(dict):
    """
    This is a generic field container.  It contains a list of potential derived
    fields, all of which know how to act on a data object and return a value.
    This object handles converting units as well as validating the availability
    of a given field.

    """
    fallback = None
    known_other_fields = ()
    known_particle_fields = ()
    extra_union_fields = ()

    def __init__(self, ds, field_list, slice_info = None):
        self._show_field_errors = []
        self.ds = ds
        # Now we start setting things up.
        self.field_list = field_list
        self.slice_info = slice_info
        self.field_aliases = {}
        self.species_names = []
        if ds is not None and is_curvilinear(ds.geometry):
            self.curvilinear = True
        else:
            self.curvilinear = False
        self.setup_fluid_aliases()

    def setup_fluid_fields(self):
        pass

    def setup_fluid_index_fields(self):
        # Now we get all our index types and set up aliases to them
        if self.ds is None: return
        index_fields = set([f for _, f in self if _ == "index"])
        for ftype in self.ds.fluid_types:
            if ftype in ("index", "deposit"): continue
            for f in index_fields:
                if (ftype, f) in self: continue
                self.alias((ftype, f), ("index", f))

    def setup_particle_fields(self, ptype, ftype='gas', num_neighbors=64):
        skip_output_units = ("code_length")
        for f, (units, aliases, dn) in sorted(self.known_particle_fields):
            units = self.ds.field_units.get((ptype, f), units)
            output_units = units
            if (f in aliases or ptype not in self.ds.particle_types_raw) and \
                units not in skip_output_units:
                u = Unit(units, registry = self.ds.unit_registry)
                if u.dimensions is not dimensionless:
                    output_units = str(self.ds.unit_system[u.dimensions])
            if (ptype, f) not in self.field_list:
                continue
            self.add_output_field((ptype, f), sampling_type="particle",
                units = units, display_name = dn, 
                output_units = output_units)
            for alias in aliases:
                self.alias((ptype, alias), (ptype, f), units = output_units)

        # We'll either have particle_position or particle_position_[xyz]
        if (ptype, "particle_position") in self.field_list or \
           (ptype, "particle_position") in self.field_aliases:
            particle_scalar_functions(ptype,
                   "particle_position", "particle_velocity",
                   self)
        else:
            # We need to check to make sure that there's a "known field" that
            # overlaps with one of the vector fields.  For instance, if we are
            # in the Stream frontend, and we have a set of scalar position
            # fields, they will overlap with -- and be overridden by -- the
            # "known" vector field that the frontend creates.  So the easiest
            # thing to do is to simply remove the on-disk field (which doesn't
            # exist) and replace it with a derived field.
            if (ptype, "particle_position") in self and \
                 self[ptype, "particle_position"]._function == NullFunc:
                self.pop((ptype, "particle_position"))
            particle_vector_functions(ptype,
                    ["particle_position_%s" % ax for ax in 'xyz'],
                    ["particle_velocity_%s" % ax for ax in 'xyz'],
                    self)
        particle_deposition_functions(ptype, "particle_position",
            "particle_mass", self)
        standard_particle_fields(self, ptype)
        # Now we check for any leftover particle fields
        for field in sorted(self.field_list):
            if field in self: continue
            if not isinstance(field, tuple):
                raise RuntimeError
            if field[0] not in self.ds.particle_types:
                continue
            self.add_output_field(field, sampling_type="particle",
                                  units = self.ds.field_units.get(field, ""))
        self.setup_smoothed_fields(ptype, 
                                   num_neighbors=num_neighbors,
                                   ftype=ftype)

    def setup_extra_union_fields(self, ptype="all"):
        if ptype != "all":
            raise RuntimeError("setup_extra_union_fields is currently" + 
                               "only enabled for particle type \"all\".")
        for units, field in self.extra_union_fields:
            add_union_field(self, ptype, field, units)

    def setup_smoothed_fields(self, ptype, num_neighbors = 64, ftype = "gas"):
        # We can in principle compute this, but it is not yet implemented.
        if (ptype, "density") not in self:
            return
        if (ptype, "smoothing_length") in self:
            sml_name = "smoothing_length"
        else:
            sml_name = None
        new_aliases = []
        for ptype2, alias_name in list(self):
            if ptype2 != ptype:
                continue
            if alias_name not in sph_whitelist_fields:
                continue
            fn = add_volume_weighted_smoothed_field(
                ptype, "particle_position", "particle_mass",
                sml_name, "density", alias_name, self,
                num_neighbors)
            if 'particle_' in alias_name:
                alias_name = alias_name.replace('particle_', '')
            new_aliases.append(((ftype, alias_name), fn[0]))
        for alias, source in new_aliases:
            #print "Aliasing %s => %s" % (alias, source)
            self.alias(alias, source)

    # Collect the names for all aliases if geometry is curvilinear
    def get_aliases_gallery(self):
        aliases_gallery = []
        known_other_fields = dict(self.known_other_fields)
        if self.curvilinear:
            for field in sorted(self.field_list):
                if field[0] in self.ds.particle_types:
                    continue
                args = known_other_fields.get(field[1], ("", [], None))
                units, aliases, display_name = args
                for alias in aliases:
                    aliases_gallery.append(alias)
        return aliases_gallery

    def setup_fluid_aliases(self, ftype='gas'):
        known_other_fields = dict(self.known_other_fields)

        # For non-Cartesian geometry, convert alias of vector fields to 
        # curvilinear coordinates
        aliases_gallery = self.get_aliases_gallery()
            
        for field in sorted(self.field_list):
            if not isinstance(field, tuple):
                raise RuntimeError
            if field[0] in self.ds.particle_types:
                continue
            args = known_other_fields.get(
                field[1], ("", [], None))
            units, aliases, display_name = args
            # We allow field_units to override this.  First we check if the
            # field *name* is in there, then the field *tuple*.
            units = self.ds.field_units.get(field[1], units)
            units = self.ds.field_units.get(field, units)
            if not isinstance(units, string_types) and args[0] != "":
                units = "((%s)*%s)" % (args[0], units)
            if isinstance(units, (numeric_type, np.number, np.ndarray)) and \
                args[0] == "" and units != 1.0:
                mylog.warning("Cannot interpret units: %s * %s, " +
                              "setting to dimensionless.", units, args[0])
                units = ""
            elif units == 1.0:
                units = ""
            self.add_output_field(field, sampling_type="cell", units = units,
                                  display_name = display_name)
            axis_names = self.ds.coordinates.axis_order
            for alias in aliases:
                if self.curvilinear: # For non-Cartesian geometry, convert vector aliases
                    
                    if alias[-2:] not in ["_x", "_y", "_z"]:
                        to_convert = False
                    else:
                        for suffix in ["x", "y", "z"]:
                            if "%s_%s" % (alias[:-2], suffix) not in aliases_gallery:
                                to_convert = False
                                break
                        to_convert = True
                    if to_convert:
                        if alias[-2:]=="_x":
                            alias = "%s_%s" % (alias[:-2], axis_names[0])
                        elif alias[-2:]=="_y":
                            alias = "%s_%s" % (alias[:-2], axis_names[1])
                        elif alias[-2:]=="_z":
                            alias = "%s_%s" % (alias[:-2], axis_names[2])
                self.alias((ftype, alias), field)

    def add_field(self, name, sampling_type, function=None, **kwargs):
        """
        Add a new field, along with supplemental metadata, to the list of
        available fields.  This respects a number of arguments, all of which
        are passed on to the constructor for
        :class:`~yt.data_objects.api.DerivedField`.

        Parameters
        ----------

        name : str
           is the name of the field.
        function : callable
           A function handle that defines the field.  Should accept
           arguments (field, data)
        units : str
           A plain text string encoding the unit.  Powers must be in
           python syntax (** instead of ^). If set to "auto" the units
           will be inferred from the return value of the field function.
        take_log : bool
           Describes whether the field should be logged
        validators : list
           A list of :class:`FieldValidator` objects
        particle_type : bool
           Is this a particle (1D) field?
        vector_field : bool
           Describes the dimensionality of the field.  Currently unused.
        display_name : str
           A name used in the plots

        """
        override = kwargs.pop("force_override", False)
        # Handle the case where the field has already been added.
        if not override and name in self:
            # See below.
            if function is None:
                def create_function(f):
                    return f
                return create_function
            return
        # add_field can be used in two different ways: it can be called
        # directly, or used as a decorator. If called directly, the
        # function will be passed in as an argument, and we simply create
        # the derived field and exit. If used as a decorator, function will
        # be None. In that case, we return a function that will be applied
        # to the function that the decorator is applied to.
        kwargs.setdefault('ds', self.ds)
        if function is None:
            def create_function(f):
                self[name] = DerivedField(name, sampling_type, f, **kwargs)
                return f
            return create_function

        if isinstance(name, tuple):
            self[name] = DerivedField(name, sampling_type, function, **kwargs)
            return

        if sampling_type == 'particle':
            ftype = 'all'
        else:
            ftype = self.ds.default_fluid_type

        if (ftype, name) not in self:
            tuple_name = (ftype, name)
            self[tuple_name] = DerivedField(tuple_name, sampling_type, function,
                                            **kwargs)
            self.alias(name, tuple_name)
        else:
            self[name] = DerivedField(name, sampling_type, function, **kwargs)

    def load_all_plugins(self, ftype="gas"):
        loaded = []
        for n in sorted(field_plugins):
            loaded += self.load_plugin(n, ftype)
            only_on_root(mylog.debug, "Loaded %s (%s new fields)",
                         n, len(loaded))
        self.find_dependencies(loaded)

    def load_plugin(self, plugin_name, ftype = "gas", skip_check = False):
        if callable(plugin_name):
            f = plugin_name
        else:
            f = field_plugins[plugin_name]
        orig = set(self.items())
        f(self, ftype, slice_info = self.slice_info)
        loaded = [n for n, v in set(self.items()).difference(orig)]
        return loaded

    def find_dependencies(self, loaded):
        deps, unavailable = self.check_derived_fields(loaded)
        self.ds.field_dependencies.update(deps)
        # Note we may have duplicated
        dfl = set(self.ds.derived_field_list).union(deps.keys())
        self.ds.derived_field_list = list(sorted(dfl, key=tupleize))
        return loaded, unavailable

    def add_output_field(self, name, sampling_type, **kwargs):
        kwargs.setdefault('ds', self.ds)
        self[name] = DerivedField(name, sampling_type, NullFunc, **kwargs)

    def alias(self, alias_name, original_name, units = None):
        if original_name not in self: return
        if units is None:
            # We default to CGS here, but in principle, this can be pluggable
            # as well.
            u = Unit(self[original_name].units,
                      registry = self.ds.unit_registry)
            if u.dimensions is not dimensionless:
                units = str(self.ds.unit_system[u.dimensions])
            else:
                units = self[original_name].units
        self.field_aliases[alias_name] = original_name
        self.add_field(alias_name,
            function = TranslationFunc(original_name),
            sampling_type = self[original_name].sampling_type,
            display_name = self[original_name].display_name,
            units = units)

    def has_key(self, key):
        # This gets used a lot
        if key in self: return True
        if self.fallback is None: return False
        return key in self.fallback

    def __missing__(self, key):
        if self.fallback is None:
            raise KeyError("No field named %s" % (key,))
        return self.fallback[key]

    @classmethod
    def create_with_fallback(cls, fallback, name = ""):
        obj = cls()
        obj.fallback = fallback
        obj.name = name
        return obj

    def __contains__(self, key):
        if dict.__contains__(self, key): return True
        if self.fallback is None: return False
        return key in self.fallback

    def __iter__(self):
        for f in dict.__iter__(self):
            yield f
        if self.fallback is not None:
            for f in self.fallback: yield f

    def keys(self):
        keys = dict.keys(self)
        if self.fallback:
            keys += list(self.fallback.keys())
        return keys

    def check_derived_fields(self, fields_to_check = None):
        deps = {}
        unavailable = []
        fields_to_check = fields_to_check or list(self.keys())
        for field in fields_to_check:
            if field not in self:
                raise YTFieldNotFound(str(field))
            fi = self[field]
            try:
                fd = fi.get_dependencies(ds = self.ds)
            except Exception as e:
                if field in self._show_field_errors:
                    raise
                if type(e) != YTFieldNotFound:
                    # if we're doing field tests, raise an error
                    # see yt.fields.tests.test_fields
                    if hasattr(self.ds, '_field_test_dataset'):
                        raise
                    mylog.debug("Raises %s during field %s detection.",
                                str(type(e)), field)
                self.pop(field)
                continue
            # This next bit checks that we can't somehow generate everything.
            # We also manually update the 'requested' attribute
            missing = not all(f in self.field_list for f in fd.requested)
            if missing:
                self.pop(field)
                unavailable.append(field)
                continue
            fd.requested = set(fd.requested)
            deps[field] = fd
            mylog.debug("Succeeded with %s (needs %s)", field, fd.requested)
        dfl = set(self.ds.derived_field_list).union(deps.keys())
        self.ds.derived_field_list = list(sorted(dfl, key=tupleize))
        return deps, unavailable



"""
Unit test for the AbsorptionSpectrum analysis module
"""

#-----------------------------------------------------------------------------
# Copyright (c) 2014, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
from yt.testing import \
    assert_allclose_units, requires_file, requires_module, \
    assert_almost_equal
from yt.analysis_modules.absorption_spectrum.absorption_line import \
    voigt_old, voigt_scipy
from yt.analysis_modules.absorption_spectrum.api import AbsorptionSpectrum
from yt.analysis_modules.cosmological_observation.api import LightRay
from yt.utilities.answer_testing.framework import \
    GenericArrayTest, \
    requires_answer_testing
import tempfile
import os
import shutil
from yt.utilities.on_demand_imports import \
    _h5py as h5
from yt.convenience import load


COSMO_PLUS = "enzo_cosmology_plus/AMRCosmology.enzo"
COSMO_PLUS_SINGLE = "enzo_cosmology_plus/RD0009/RD0009"
GIZMO_PLUS = "gizmo_cosmology_plus/N128L16.param"
GIZMO_PLUS_SINGLE = "gizmo_cosmology_plus/snap_N128L16_151.hdf5"
ISO_GALAXY = "IsolatedGalaxy/galaxy0030/galaxy0030"
FIRE = "FIRE_M12i_ref11/snapshot_600.hdf5"

@requires_file(COSMO_PLUS)
@requires_answer_testing()
def test_absorption_spectrum_cosmo():
    """
    This test generates an absorption spectrum from a compound light ray on a
    grid dataset
    """
    # Set up in a temp dir
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    lr = LightRay(COSMO_PLUS, 'Enzo', 0.0, 0.03)

    lr.make_light_ray(seed=1234567,
                      fields=['temperature', 'density', 'H_number_density'],
                      data_filename='lightray.h5')

    sp = AbsorptionSpectrum(900.0, 1800.0, 10000)

    my_label = 'HI Lya'
    field = 'H_number_density'
    wavelength = 1215.6700  # Angstroms
    f_value = 4.164E-01
    gamma = 6.265e+08
    mass = 1.00794

    sp.add_line(my_label, field, wavelength, f_value,
                gamma, mass, label_threshold=1.e10)

    my_label = 'HI Lya'
    field = 'H_number_density'
    wavelength = 912.323660  # Angstroms
    normalization = 1.6e17
    index = 3.0

    sp.add_continuum(my_label, field, wavelength, normalization, index)

    wavelength, flux = sp.make_spectrum('lightray.h5',
                                        output_file='spectrum.h5',
                                        line_list_file='lines.txt',
                                        use_peculiar_velocity=True)

    # load just-generated hdf5 file of spectral data (for consistency)
    data = h5.File('spectrum.h5', 'r')

    for key in data.keys():
        func = lambda x=key: data[x][:]
        func.__name__ = "{}_cosmo".format(key)
        test = GenericArrayTest(None, func)
        test_absorption_spectrum_cosmo.__name__ = test.description
        yield test

    # clean up
    os.chdir(curdir)
    shutil.rmtree(tmpdir)

@requires_file(COSMO_PLUS_SINGLE)
@requires_answer_testing()
def test_absorption_spectrum_non_cosmo():
    """
    This test generates an absorption spectrum from a simple light ray on a
    grid dataset
    """

    # Set up in a temp dir
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    lr = LightRay(COSMO_PLUS_SINGLE)

    ray_start = [0,0,0]
    ray_end = [1,1,1]
    lr.make_light_ray(start_position=ray_start, end_position=ray_end,
                      fields=['temperature', 'density', 'H_number_density'],
                      data_filename='lightray.h5')

    sp = AbsorptionSpectrum(1200.0, 1300.0, 10001)

    my_label = 'HI Lya'
    field = 'H_number_density'
    wavelength = 1215.6700  # Angstroms
    f_value = 4.164E-01
    gamma = 6.265e+08
    mass = 1.00794

    sp.add_line(my_label, field, wavelength, f_value,
                gamma, mass, label_threshold=1.e10)

    wavelength, flux = sp.make_spectrum('lightray.h5',
                                        output_file='spectrum.h5',
                                        line_list_file='lines.txt',
                                        use_peculiar_velocity=True)

    # load just-generated hdf5 file of spectral data (for consistency)
    data = h5.File('spectrum.h5', 'r')
    
    for key in data.keys():
        func = lambda x=key: data[x][:]
        func.__name__ = "{}_non_cosmo".format(key)
        test = GenericArrayTest(None, func)
        test_absorption_spectrum_non_cosmo.__name__ = test.description
        yield test

    # clean up
    os.chdir(curdir)
    shutil.rmtree(tmpdir)

@requires_file(COSMO_PLUS_SINGLE)
@requires_answer_testing()
def test_absorption_spectrum_non_cosmo_novpec():
    """
    This test generates an absorption spectrum from a simple light ray on a
    grid dataset
    """

    # Set up in a temp dir
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    lr = LightRay(COSMO_PLUS_SINGLE)

    ray_start = [0,0,0]
    ray_end = [1,1,1]
    lr.make_light_ray(start_position=ray_start, end_position=ray_end,
                      fields=['temperature', 'density', 'H_number_density'],
                      data_filename='lightray.h5', use_peculiar_velocity=False)

    sp = AbsorptionSpectrum(1200.0, 1300.0, 10001)

    my_label = 'HI Lya'
    field = 'H_number_density'
    wavelength = 1215.6700  # Angstroms
    f_value = 4.164E-01
    gamma = 6.265e+08
    mass = 1.00794

    sp.add_line(my_label, field, wavelength, f_value,
                gamma, mass, label_threshold=1.e10)

    wavelength, flux = sp.make_spectrum('lightray.h5',
                                        output_file='spectrum.h5',
                                        line_list_file='lines.txt',
                                        use_peculiar_velocity=False)

    # load just-generated hdf5 file of spectral data (for consistency)
    data = h5.File('spectrum.h5', 'r')

    for key in data.keys():
        func = lambda x=key: data[x][:]
        func.__name__ = "{}_non_cosmo_novpec".format(key)
        test = GenericArrayTest(None, func)
        test_absorption_spectrum_non_cosmo_novpec.__name__ = test.description
        yield test

    # clean up
    os.chdir(curdir)
    shutil.rmtree(tmpdir)

@requires_file(COSMO_PLUS_SINGLE)
def test_equivalent_width_conserved():
    """
    This tests that the equivalent width of the optical depth is conserved 
    regardless of the bin width employed in wavelength space.
    Unresolved lines should still deposit optical depth into the spectrum.
    """

    # Set up in a temp dir
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    lr = LightRay(COSMO_PLUS_SINGLE)

    ray_start = [0,0,0]
    ray_end = [1,1,1]
    lr.make_light_ray(start_position=ray_start, end_position=ray_end,
                      fields=['temperature', 'density', 'H_number_density'],
                      data_filename='lightray.h5')

    my_label = 'HI Lya'
    field = 'H_number_density'
    wave = 1215.6700  # Angstroms
    f_value = 4.164E-01
    gamma = 6.265e+08
    mass = 1.00794

    lambda_min= 1200
    lambda_max= 1300
    lambda_bin_widths = [1e-3, 1e-2, 1e-1, 1e0, 1e1]
    total_tau = []

    for lambda_bin_width in lambda_bin_widths:
        n_lambda = ((lambda_max - lambda_min)/ lambda_bin_width) + 1
        sp = AbsorptionSpectrum(lambda_min=lambda_min, lambda_max=lambda_max, 
                                n_lambda=n_lambda)
        sp.add_line(my_label, field, wave, f_value, gamma, mass)
        wavelength, flux = sp.make_spectrum('lightray.h5')
        total_tau.append((lambda_bin_width * sp.tau_field).sum())
        
    # assure that the total tau values are all within 1e-3 of each other
    for tau in total_tau:
        assert_almost_equal(tau, total_tau[0], 3)

    # clean up
    os.chdir(curdir)
    shutil.rmtree(tmpdir)


@requires_file(COSMO_PLUS_SINGLE)
@requires_module("astropy")
def test_absorption_spectrum_fits():
    """
    This test generates an absorption spectrum and saves it as a fits file.
    """

    # Set up in a temp dir
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    lr = LightRay(COSMO_PLUS_SINGLE)

    ray_start = [0,0,0]
    ray_end = [1,1,1]
    lr.make_light_ray(start_position=ray_start, end_position=ray_end,
                      fields=['temperature', 'density', 'H_number_density'],
                      data_filename='lightray.h5')

    sp = AbsorptionSpectrum(900.0, 1800.0, 10000)

    my_label = 'HI Lya'
    field = 'H_number_density'
    wavelength = 1215.6700  # Angstroms
    f_value = 4.164E-01
    gamma = 6.265e+08
    mass = 1.00794

    sp.add_line(my_label, field, wavelength, f_value,
                gamma, mass, label_threshold=1.e10)

    my_label = 'HI Lya'
    field = 'H_number_density'
    wavelength = 912.323660  # Angstroms
    normalization = 1.6e17
    index = 3.0

    sp.add_continuum(my_label, field, wavelength, normalization, index)

    wavelength, flux = sp.make_spectrum('lightray.h5',
                                        output_file='spectrum.fits',
                                        line_list_file='lines.txt',
                                        use_peculiar_velocity=True)

    # clean up
    os.chdir(curdir)
    shutil.rmtree(tmpdir)


@requires_module("scipy")
def test_voigt_profiles():
    a = 1.7e-4
    x = np.linspace(5.0, -3.6, 60)
    assert_allclose_units(voigt_old(a, x), voigt_scipy(a, x), 1e-8)

@requires_file(GIZMO_PLUS)
@requires_answer_testing()
def test_absorption_spectrum_cosmo_sph():
    """
    This test generates an absorption spectrum from a compound light ray on a
    particle dataset
    """
    # Set up in a temp dir
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    lr = LightRay(GIZMO_PLUS, 'Gadget', 0.0, 0.01)

    lr.make_light_ray(seed=1234567,
                      fields=[('gas', 'temperature'), 
                              ('gas', 'H_number_density')],
                      data_filename='lightray.h5')

    sp = AbsorptionSpectrum(900.0, 1800.0, 10000)

    my_label = 'HI Lya'
    field = ('gas', 'H_number_density')
    wavelength = 1215.6700  # Angstroms
    f_value = 4.164E-01
    gamma = 6.265e+08
    mass = 1.00794

    sp.add_line(my_label, field, wavelength, f_value,
                gamma, mass, label_threshold=1.e10)

    my_label = 'HI Lya'
    field = ('gas', 'H_number_density')
    wavelength = 912.323660  # Angstroms
    normalization = 1.6e17
    index = 3.0

    sp.add_continuum(my_label, field, wavelength, normalization, index)

    wavelength, flux = sp.make_spectrum('lightray.h5',
                                        output_file='spectrum.h5',
                                        line_list_file='lines.txt',
                                        use_peculiar_velocity=True)

    # load just-generated hdf5 file of spectral data (for consistency)
    data = h5.File('spectrum.h5', 'r')

    for key in data.keys():
        func = lambda x=key: data[x][:]
        func.__name__ = "{}_cosmo_sph".format(key)
        test = GenericArrayTest(None, func)
        test_absorption_spectrum_cosmo_sph.__name__ = test.description
        yield test

    # clean up
    os.chdir(curdir)
    shutil.rmtree(tmpdir)

@requires_file(GIZMO_PLUS_SINGLE)
@requires_answer_testing()
def test_absorption_spectrum_non_cosmo_sph():
    """
    This test generates an absorption spectrum from a simple light ray on a
    particle dataset
    """

    # Set up in a temp dir
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    ds = load(GIZMO_PLUS_SINGLE)
    lr = LightRay(ds)
    ray_start = ds.domain_left_edge
    ray_end = ds.domain_right_edge
    lr.make_light_ray(start_position=ray_start, end_position=ray_end,
                      fields=[('gas', 'temperature'), 
                              ('gas', 'H_number_density')],
                      data_filename='lightray.h5')

    sp = AbsorptionSpectrum(1200.0, 1300.0, 10001)

    my_label = 'HI Lya'
    field = ('gas', 'H_number_density')
    wavelength = 1215.6700  # Angstroms
    f_value = 4.164E-01
    gamma = 6.265e+08
    mass = 1.00794

    sp.add_line(my_label, field, wavelength, f_value,
                gamma, mass, label_threshold=1.e10)

    wavelength, flux = sp.make_spectrum('lightray.h5',
                                        output_file='spectrum.h5',
                                        line_list_file='lines.txt',
                                        use_peculiar_velocity=True)

    # load just-generated hdf5 file of spectral data (for consistency)
    data = h5.File('spectrum.h5', 'r')
    
    for key in data.keys():
        func = lambda x=key: data[x][:]
        func.__name__ = "{}_non_cosmo_sph".format(key)
        test = GenericArrayTest(None, func)
        test_absorption_spectrum_non_cosmo_sph.__name__ = test.description
        yield test

    # clean up
    os.chdir(curdir)
    shutil.rmtree(tmpdir)

@requires_file(ISO_GALAXY)
@requires_answer_testing()
def test_absorption_spectrum_with_continuum():
    """
    This test generates an absorption spectrum from a simple light ray on a
    grid dataset and adds Lyman alpha and Lyman continuum to it
    """

    # Set up in a temp dir
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    ds = load(ISO_GALAXY)
    lr = LightRay(ds)

    ray_start = ds.domain_left_edge
    ray_end = ds.domain_right_edge
    lr.make_light_ray(start_position=ray_start, end_position=ray_end,
                      fields=['temperature', 'density', 'H_number_density'],
                      data_filename='lightray.h5')

    sp = AbsorptionSpectrum(800.0, 1300.0, 5001)

    my_label = 'HI Lya'
    field = 'H_number_density'
    wavelength = 1215.6700  # Angstroms
    f_value = 4.164E-01
    gamma = 6.265e+08
    mass = 1.00794

    sp.add_line(my_label, field, wavelength, f_value,
                gamma, mass, label_threshold=1.e10)

    my_label = 'Ly C'
    field = 'H_number_density'
    wavelength = 912.323660  # Angstroms
    normalization = 1.6e17
    index = 3.0

    sp.add_continuum(my_label, field, wavelength, normalization, index)

    wavelength, flux = sp.make_spectrum('lightray.h5',
                                        output_file='spectrum.h5',
                                        line_list_file='lines.txt',
                                        use_peculiar_velocity=True)

    # load just-generated hdf5 file of spectral data (for consistency)
    data = h5.File('spectrum.h5', 'r')
    
    for key in data.keys():
        func = lambda x=key: data[x][:]
        func.__name__ = "{}_continuum".format(key)
        test = GenericArrayTest(None, func)
        test_absorption_spectrum_with_continuum.__name__ = test.description
        yield test

    # clean up
    os.chdir(curdir)
    shutil.rmtree(tmpdir)

@requires_file(FIRE)
def test_absorption_spectrum_with_zero_field():
    """
    This test generates an absorption spectrum with some 
    particle dataset
    """

    # Set up in a temp dir
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    ds = load(FIRE)
    lr = LightRay(ds)

    # Define species and associated parameters to add to continuum
    # Parameters used for both adding the transition to the spectrum
    # and for fitting
    # Note that for single species that produce multiple lines
    # (as in the OVI doublet), 'numLines' will be equal to the number
    # of lines, and f,gamma, and wavelength will have multiple values.

    HI_parameters = {
        'name': 'HI',
        'field': 'H_number_density',
        'f': [.4164],
        'Gamma': [6.265E8],
        'wavelength': [1215.67],
        'mass': 1.00794,
        'numLines': 1,
        'maxN': 1E22, 'minN': 1E11,
        'maxb': 300, 'minb': 1,
        'maxz': 6, 'minz': 0,
        'init_b': 30,
        'init_N': 1E14
    }

    species_dicts = {'HI': HI_parameters}


    # Get all fields that need to be added to the light ray
    fields = [('gas','temperature')]
    for s, params in species_dicts.items():
        fields.append(params['field'])

    # With a single dataset, a start_position and
    # end_position or trajectory must be given.
    # Trajectory should be given as (r, theta, phi)
    lr.make_light_ray(
        start_position=ds.arr([0., 0., 0.], 'unitary'),
        end_position=ds.arr([1., 1., 1.], 'unitary'),
        solution_filename='test_lightraysolution.txt',
        data_filename='test_lightray.h5',
        fields=fields)
    
    # Create an AbsorptionSpectrum object extending from
    # lambda = 900 to lambda = 1800, with 10000 pixels
    sp = AbsorptionSpectrum(900.0, 1400.0, 50000)
    
    # Iterate over species
    for s, params in species_dicts.items():
        # Iterate over transitions for a single species
        for i in range(params['numLines']):
            # Add the lines to the spectrum
            sp.add_line(
                s, params['field'],
                params['wavelength'][i], params['f'][i],
                params['Gamma'][i], params['mass'],
                label_threshold=1.e10)
    
    
    # Make and save spectrum
    wavelength, flux = sp.make_spectrum(
        'test_lightray.h5',
        output_file='test_spectrum.h5',
        line_list_file='test_lines.txt',
        use_peculiar_velocity=True)

    # clean up
    os.chdir(curdir)
    shutil.rmtree(tmpdir)



"""
RAMSES-specific fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import os
import numpy as np

from yt import units
from yt.fields.field_detector import \
    FieldDetector
from yt.fields.field_info_container import \
    FieldInfoContainer
from yt.frontends.ramses.io import \
    convert_ramses_ages
from yt.funcs import \
    issue_deprecation_warning
from yt.utilities.physical_constants import \
    boltzmann_constant_cgs, \
    mass_hydrogen_cgs, \
    mh
from yt.utilities.linear_interpolators import \
    BilinearFieldInterpolator
from yt.utilities.cython_fortran_utils import FortranFile
from .field_handlers import RTFieldFileHandler

from yt.utilities.physical_constants import \
    mp
from yt.funcs import mylog

b_units = "code_magnetic"
ra_units = "code_length / code_time**2"
rho_units = "code_density"
vel_units = "code_velocity"
pressure_units = "code_pressure"
ener_units = "code_mass * code_velocity**2"
ang_mom_units = "code_mass * code_velocity * code_length"
cooling_function_units=" erg * cm**3 /s"
cooling_function_prime_units=" erg * cm**3 /s/K"
flux_unit = "1 / code_length**2 / code_time"
number_density_unit = "1 / code_length**3"

known_species_masses = dict(
  (sp, mh * v) for sp, v in [
                ("HI", 1.0),
                ("HII", 1.0),
                ("Electron", 1.0),
                ("HeI", 4.0),
                ("HeII", 4.0),
                ("HeIII", 4.0),
                ("H2I", 2.0),
                ("H2II", 2.0),
                ("HM", 1.0),
                ("DI", 2.0),
                ("DII", 2.0),
                ("HDI", 3.0),
    ])

_cool_axes = ("lognH", "logT")#, "logTeq")
_cool_arrs = (("cooling_primordial", cooling_function_units),
              ("heating_primordial", cooling_function_units),
              ("cooling_compton",cooling_function_units),
              ("heating_compton",cooling_function_units),
              ("cooling_metal", cooling_function_units),
              ("cooling_primordial_prime", cooling_function_prime_units),
              ("heating_primordial_prime", cooling_function_prime_units),
              ("cooling_compton_prime", cooling_function_prime_units),
              ("heating_compton_prime", cooling_function_prime_units),
              ("cooling_metal_prime", cooling_function_prime_units),
              ("mu", None),
              ("abundances", None))
_cool_species = ("Electron_number_density",
                 "HI_number_density",
                 "HII_number_density",
                 "HeI_number_density",
                 "HeII_number_density",
                 "HeIII_number_density")

_X = 0.76 # H fraction, hardcoded
_Y = 0.24 # He fraction, hardcoded


class RAMSESFieldInfo(FieldInfoContainer):
    known_other_fields = (
        ("Density", (rho_units, ["density"], None)),
        ("x-velocity", (vel_units, ["velocity_x"], None)),
        ("y-velocity", (vel_units, ["velocity_y"], None)),
        ("z-velocity", (vel_units, ["velocity_z"], None)),
        ("Pres_IR", (pressure_units, ["pres_IR", "pressure_IR"], None)),
        ("Pressure", (pressure_units, ["pressure"], None)),
        ("Metallicity", ("", ["metallicity"], None)),
        ("HII",  ("", ["H_p1_fraction"], None)),
        ("HeII", ("", ["He_p1_fraction"], None)),
        ("HeIII",("", ["He_p2_fraction"], None)),
        ("x-acceleration", (ra_units, ['acceleration_x'], None)),
        ("y-acceleration", (ra_units, ['acceleration_y'], None)),
        ("z-acceleration", (ra_units, ['acceleration_z'], None)),
        ("Potential", (ener_units, ['potential'], None)),
        ("B_x_left", (b_units, ['magnetic_field_x_left'],None)),
        ("B_x_right", (b_units, ['magnetic_field_x_right'],None)),
        ("B_y_left", (b_units, ['magnetic_field_y_left'],None)),
        ("B_y_right", (b_units, ['magnetic_field_y_right'],None)),
        ("B_z_left", (b_units, ['magnetic_field_z_left'],None)),
        ("B_z_right", (b_units, ['magnetic_field_z_right'],None)),
    )
    known_particle_fields = (
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_position_z", ("code_length", [], None)),
        ("particle_velocity_x", (vel_units, [], None)),
        ("particle_velocity_y", (vel_units, [], None)),
        ("particle_velocity_z", (vel_units, [], None)),
        ("particle_mass", ("code_mass", [], None)),
        ("particle_identity", ("", ["particle_index"], None)),
        ("particle_refinement_level", ("", [], None)),
        ("particle_birth_time", ("code_time", ['age'], None)),
        ("conformal_birth_time", ("", [], None)),
        ("particle_metallicity", ("", [], None)),
        ("particle_family", ("", [], None)),
        ("particle_tag", ("", [], None))
    )

    known_sink_fields = (
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_position_z", ("code_length", [], None)),
        ("particle_velocity_x", (vel_units, [], None)),
        ("particle_velocity_y", (vel_units, [], None)),
        ("particle_velocity_z", (vel_units, [], None)),
        ("particle_mass", ("code_mass", [], None)),
        ("particle_identifier", ("", ["particle_index"], None)),
        ("particle_birth_time", ("code_time", ['age'], None)),
        ("BH_real_accretion", ("code_mass/code_time", [], None)),
        ("BH_bondi_accretion", ("code_mass/code_time", [], None)),
        ("BH_eddington_accretion", ("code_mass/code_time", [], None)),
        ("BH_esave", (ener_units, [], None)),
        ("gas_spin_x", (ang_mom_units, [], None)),
        ("gas_spin_y", (ang_mom_units, [], None)),
        ("gas_spin_z", (ang_mom_units, [], None)),
        ("BH_spin_x", ("", [], None)),
        ("BH_spin_y", ("", [], None)),
        ("BH_spin_z", ("", [], None)),
        ("BH_spin", (ang_mom_units, [], None)),
        ("BH_efficiency", ("", [], None))
    )

    def setup_particle_fields(self, ptype):
        super(RAMSESFieldInfo, self).setup_particle_fields(ptype)
        def particle_age(field, data):
            msg = ("The RAMSES particle_age field has been deprecated since "
                   "it did not actually represent particle ages in all "
                   "cases. To get the time when a particle was formed use "
                   "the particle_birth_time field instead. To get the "
                   "age of a star particle, use the star_age field")
            if not isinstance(data, FieldDetector):
                issue_deprecation_warning(msg, stacklevel=2)
            if data.ds.cosmological_simulation:
                conformal_age = data[ptype, 'conformal_birth_time']
                ret = convert_ramses_ages(data.ds, conformal_age)
                return data.ds.arr(ret, 'code_time')
            else:
                return data[ptype, 'particle_birth_time']
        self.add_field((ptype, 'particle_age'), sampling_type='particle',
                       function=particle_age,
                       units=self.ds.unit_system['time'])
        def star_age(field, data):
            if data.ds.cosmological_simulation:
                conformal_age = data[ptype, 'conformal_birth_time']
                formation_time = convert_ramses_ages(data.ds, conformal_age)
                formation_time = data.ds.arr(formation_time, 'code_time')
            else:
                formation_time = data['particle_birth_time']
            return data.ds.current_time - formation_time
        self.add_field((ptype, 'star_age'), sampling_type='particle',
                       function=star_age, units=self.ds.unit_system['time'])

    def setup_fluid_fields(self):
        def _temperature(field, data):
            rv = data["gas", "pressure"]/data["gas", "density"]
            rv *= mass_hydrogen_cgs/boltzmann_constant_cgs
            return rv
        self.add_field(("gas", "temperature"), sampling_type="cell",  function=_temperature,
                        units=self.ds.unit_system["temperature"])
        self.create_cooling_fields()

        # See if we need to load the rt fields
        rt_flag = RTFieldFileHandler.any_exist(self.ds)
        if rt_flag: # rt run
            self.create_rt_fields()

        #Load magnetic fields
        if ('gas','magnetic_field_x_left') in self:
            self.create_magnetic_fields()

    def create_magnetic_fields(self):
        #Calculate cell-centred magnetic fields from face-centred
        def mag_field(ax):
            def _mag_field(field, data):
                return (data['magnetic_field_%s_left' % ax]+data['magnetic_field_%s_right' % ax])/2
            return _mag_field

        for ax in self.ds.coordinates.axis_order:
            self.add_field(('gas',"magnetic_field_%s" % ax),
                           sampling_type='cell',
                           function=mag_field(ax),
                           units=self.ds.unit_system['magnetic_field_cgs'])

        def _divB(field,data):
            '''Calculate magnetic field divergence'''
            out = np.zeros_like(data['magnetic_field_x_right'])
            for ax in data.ds.coordinates.axis_order:
                out += data['magnetic_field_%s_right' % ax] - data['magnetic_field_%s_left' % ax] 
            return out/data['dx']

        self.add_field(('gas','magnetic_field_divergence'),
                       sampling_type='cell',
                       function=_divB,
                       units=self.ds.unit_system['magnetic_field_cgs']/ self.ds.unit_system['length'])

    def create_rt_fields(self):
        self.ds.fluid_types += ('rt', )
        p = RTFieldFileHandler.get_rt_parameters(self.ds).copy()
        p.update(self.ds.parameters)
        ngroups = p['nGroups']
        rt_c = p['rt_c_frac'] * units.c / (p['unit_l'] / p['unit_t'])
        dens_conv = (p['unit_np'] / rt_c).value / units.cm**3

        ########################################
        # Adding the fields in the hydro_* files
        def _temp_IR(field, data):
            rv = data["gas", "pres_IR"]/data["gas", "density"]
            rv *= mass_hydrogen_cgs/boltzmann_constant_cgs
            return rv
        self.add_field(("gas", "temp_IR"), sampling_type="cell",
                       function=_temp_IR,
                       units=self.ds.unit_system["temperature"])
        for species in ['H_p1', 'He_p1', 'He_p2']:
            def _species_density(field, data):
                return data['gas', species+'_fraction']*data['gas', 'density']
            self.add_field(('gas', species+'_density'), sampling_type='cell',
                           function=_species_density,
                           units=self.ds.unit_system['density'])
            def _species_mass(field, data):
                return (data['gas', species+'_density']*
                        data['index', 'cell_volume'])
            self.add_field(('gas', species+'_mass'), sampling_type='cell',
                           function=_species_mass,
                           units=self.ds.unit_system['mass'])

        ########################################
        # Adding the fields in the rt_ files
        def gen_pdens(igroup):
            def _photon_density(field, data):
                rv = data['ramses-rt', 'Photon_density_%s' % (igroup + 1)] * dens_conv
                return rv
            return _photon_density

        for igroup in range(ngroups):
            self.add_field(('rt', 'photon_density_%s' % (igroup + 1)), sampling_type='cell',
                           function=gen_pdens(igroup),
                           units=self.ds.unit_system['number_density'])

        flux_conv = p['unit_pf'] / units.cm**2 / units.s

        def gen_flux(key, igroup):
            def _photon_flux(field, data):
                rv = data['ramses-rt', 'Photon_flux_%s_%s' % (key, igroup + 1)] * flux_conv
                return rv
            return _photon_flux

        flux_unit = str(1/self.ds.unit_system['time']/self.ds.unit_system['length']**2)
        for key in 'xyz':
            for igroup in range(ngroups):
                self.add_field(('rt', 'photon_flux_%s_%s' % (key, igroup + 1)), sampling_type='cell',
                               function=gen_flux(key, igroup),
                               units=flux_unit)


    def create_cooling_fields(self):
        num = os.path.basename(self.ds.parameter_filename).split("."
                )[0].split("_")[1]
        filename = "%s/cooling_%05i.out" % (
            os.path.dirname(self.ds.parameter_filename), int(num))

        if not os.path.exists(filename):
            mylog.warning('This output has no cooling fields')
            return

        #Function to create the cooling fields
        def _create_field(name, interp_object,unit):
            def _func(field, data):
                shape = data["temperature"].shape
                d = {'lognH': np.log10(_X*data["density"]/mh).ravel(),
                     'logT' : np.log10(data["temperature"]).ravel()}
                rv = interp_object(d).reshape(shape)
                if name[-1] != 'mu':
                    rv = 10**interp_object(d).reshape(shape)
                cool=data.ds.arr(rv, unit)
                if 'metal' in name[-1].split('_'):
                    cool=cool*data['metallicity']/0.02  #Ramses uses Zsolar=0.02
                elif 'compton' in name[-1].split('_'):
                    cool=data.ds.arr(rv, unit+'/cm**3')
                    cool=cool/data['number_density']    #Compton cooling/heating is written to file in erg/s
                return cool
            self.add_field(name=name, sampling_type="cell", function=_func,
                           units=unit)

        #Load cooling files
        avals = {}
        tvals = {}
        with FortranFile(filename) as fd:
            n1, n2 = fd.read_vector('i')
            for ax in _cool_axes:
                avals[ax] = fd.read_vector('d')
            for i,(tname, unit) in enumerate(_cool_arrs):
                var = fd.read_vector('d')
                if var.size==n1 and i==0:
                    #If this case occurs, the cooling files were produced pre-2010 in a format
                    #that is no longer supported
                    mylog.warning('This cooling file format is no longer supported. Cooling field loading skipped.')
                    return
                if var.size == n1*n2:
                    tvals[tname] = dict(data=var.reshape((n1, n2), order='F'), unit=unit)
                else:
                    var = var.reshape((n1, n2, var.size // (n1*n2)), order='F')
                    for i in range(var.shape[-1]):
                        tvals[_cool_species[i]] = dict(data=var[:,:,i], unit="1/cm**3")

        #Add the mu field first, as it is needed for the number density
        interp = BilinearFieldInterpolator(tvals['mu']['data'],
                                           (avals["lognH"], avals["logT"]),
                                           ["lognH", "logT"], truncate = True)
        _create_field(("gas", 'mu'), interp, tvals['mu']['unit'])

        #Add the number density field, based on mu
        def _number_density(field,data):
            return data[('gas','density')]/mp/data['mu']
        self.add_field(name=('gas','number_density'),sampling_type="cell",function=_number_density,
                       units=number_density_unit)

        #Add the cooling and heating fields, which need the number density field
        for key in tvals:
            if key != 'mu':
                interp = BilinearFieldInterpolator(tvals[key]['data'],
                                                   (avals["lognH"], avals["logT"]),
                                                   ["lognH", "logT"], truncate = True)
                _create_field(("gas", key), interp, tvals[key]['unit'])

        #Add total cooling and heating fields
        def _all_cool(field,data):
            return data['cooling_primordial']+data['cooling_metal']+data['cooling_compton']
        def _all_heat(field,data):
            return data['heating_primordial']+data['heating_compton']

        self.add_field(name=('gas','cooling_total'),sampling_type="cell",function=_all_cool,
                       units=cooling_function_units)
        self.add_field(name=('gas','heating_total'),sampling_type="cell",function=_all_heat,
                       units=cooling_function_units)

        #Add net cooling fields
        def _net_cool(field,data):
            return data['cooling_total']-data['heating_total']

        self.add_field(name=('gas','cooling_net'),sampling_type="cell",function=_net_cool,
                       units=cooling_function_units)



"""
Field-related exceptions.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


class ValidationException(Exception):
    pass

class NeedsGridType(ValidationException):
    def __init__(self, ghost_zones = 0, fields=None):
        self.ghost_zones = ghost_zones
        self.fields = fields
    def __str__(self):
        return "(%s, %s)" % (self.ghost_zones, self.fields)

class NeedsOriginalGrid(NeedsGridType):
    def __init__(self):
        self.ghost_zones = 0

class NeedsDataField(ValidationException):
    def __init__(self, missing_fields):
        self.missing_fields = missing_fields
    def __str__(self):
        return "(%s)" % (self.missing_fields)

class NeedsProperty(ValidationException):
    def __init__(self, missing_properties):
        self.missing_properties = missing_properties
    def __str__(self):
        return "(%s)" % (self.missing_properties)

class NeedsParameter(ValidationException):
    def __init__(self, missing_parameters):
        self.missing_parameters = missing_parameters
    def __str__(self):
        return "(%s)" % (self.missing_parameters)

class NeedsConfiguration(ValidationException):
    def __init__(self, parameter, value):
        self.parameter = parameter
        self.value = value
    def __str__(self):
        return "(Needs %s = %s)" % (self.parameter, self.value)

class FieldUnitsError(Exception):
    pass




"""
Clump finding helper classes



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
import uuid

from yt.fields.derived_field import \
    ValidateSpatial
from yt.frontends.ytdata.utilities import \
    save_as_dataset
from yt.funcs import \
    issue_deprecation_warning, \
    get_output_filename, \
    mylog
from yt.utilities.tree_container import \
    TreeContainer

from .clump_info_items import \
    clump_info_registry
from .clump_validators import \
    clump_validator_registry
from .contour_finder import \
    identify_contours

def add_contour_field(ds, contour_key):
    def _contours(field, data):
        fd = data.get_field_parameter("contour_slices_%s" % contour_key)
        vals = data["index", "ones"] * -1
        if fd is None or fd == 0.0:
            return vals
        for sl, v in fd.get(data.id, []):
            vals[sl] = v
        return vals

    ds.add_field(("index", "contours_%s" % contour_key),
                 function=_contours,
                 validators=[ValidateSpatial(0)],
                 take_log=False,
                 display_field=False,
                 sampling_type='cell',
                 units='')

class Clump(TreeContainer):
    def __init__(self, data, field, parent=None,
                 clump_info=None, validators=None,
                 base=None, contour_key=None,
                 contour_id=None):
        self.data = data
        self.field = field
        self.parent = parent
        self.quantities = data.quantities
        self.min_val = self.data[field].min()
        self.max_val = self.data[field].max()
        self.info = {}
        self.children = []

        # is this the parent clump?
        if base is None:
            base = self
            self.total_clumps = 0

        if clump_info is None:
            self.set_default_clump_info()
        else:
            self.clump_info = clump_info

        for ci in self.clump_info:
            ci(self)

        self.base = base
        self.clump_id = self.base.total_clumps
        self.base.total_clumps += 1
        self.contour_key = contour_key
        self.contour_id = contour_id

        if parent is not None:
            self.data.parent = self.parent.data

        if validators is None:
            validators = []
        self.validators = validators
        # Return value of validity function.
        self.valid = None

    _leaves = None
    @property
    def leaves(self):
        if self._leaves is not None:
            return self._leaves

        self._leaves = []
        for clump in self:
            if not clump.children:
                self._leaves.append(clump)
        return self._leaves

    def add_validator(self, validator, *args, **kwargs):
        """
        Add a validating function to determine whether the clump should 
        be kept.
        """
        callback = clump_validator_registry.find(validator, *args, **kwargs)
        self.validators.append(callback)
        for child in self.children:
            child.add_validator(validator)

    def add_info_item(self, info_item, *args, **kwargs):
        "Adds an entry to clump_info list and tells children to do the same."

        callback = clump_info_registry.find(info_item, *args, **kwargs)
        callback(self)
        self.clump_info.append(callback)
        for child in self.children:
            child.add_info_item(info_item)

    def set_default_clump_info(self):
        "Defines default entries in the clump_info array."

        # add_info_item is recursive so this function does not need to be.
        self.clump_info = []

        self.add_info_item("total_cells")
        self.add_info_item("cell_mass")

        if any("jeans" in f for f in self.data.pf.field_list):
            self.add_info_item("mass_weighted_jeans_mass")
            self.add_info_item("volume_weighted_jeans_mass")

        self.add_info_item("max_grid_level")

        if any("number_density" in f for f in self.data.pf.field_list):
            self.add_info_item("min_number_density")
            self.add_info_item("max_number_density")

    def clear_clump_info(self):
        """
        Clears the clump_info array and passes the instruction to its 
        children.
        """

        self.clump_info = []
        for child in self.children:
            child.clear_clump_info()

    def find_children(self, min_val, max_val = None):
        if self.children:
            mylog.info("Wiping out existing children clumps: %d.",
                       len(self.children))
        self.children = []
        if max_val is None: max_val = self.max_val
        nj, cids = identify_contours(self.data, self.field, min_val, max_val)
        # Here, cids is the set of slices and values, keyed by the
        # parent_grid_id, that defines the contours.  So we can figure out all
        # the unique values of the contours by examining the list here.
        unique_contours = set([])
        for sl_list in cids.values():
            for sl, ff in sl_list:
                unique_contours.update(np.unique(ff))
        contour_key = uuid.uuid4().hex
        base_object = getattr(self.data, 'base_object', self.data)
        add_contour_field(base_object.ds, contour_key)
        for cid in sorted(unique_contours):
            if cid == -1: continue
            new_clump = base_object.cut_region(
                    ["obj['contours_%s'] == %s" % (contour_key, cid)],
                    {('contour_slices_%s' % contour_key): cids})
            if new_clump["ones"].size == 0:
                # This is to skip possibly duplicate clumps.
                # Using "ones" here will speed things up.
                continue
            self.children.append(Clump(new_clump, self.field, parent=self,
                                       validators=self.validators,
                                       base=self.base,
                                       clump_info=self.clump_info,
                                       contour_key=contour_key,
                                       contour_id=cid))

    def __iter__(self):
        yield self
        for child in self.children:
            for a_node in child:
                yield a_node

    def save_as_dataset(self, filename=None, fields=None):
        r"""Export clump tree to a reloadable yt dataset.
        This function will take a clump object and output a dataset
        containing the fields given in the ``fields`` list and all info
        items.  The resulting dataset can be reloaded as a yt dataset.

        Parameters
        ----------
        filename : str, optional
            The name of the file to be written.  If None, the name
            will be a combination of the original dataset and the clump
            index.
        fields : list of strings or tuples, optional
            If this is supplied, it is the list of fields to be saved to
            disk.

        Returns
        -------
        filename : str
            The name of the file that has been created.

        Examples
        --------

        >>> import yt
        >>> from yt.data_objects.level_sets.api import \
        ...         Clump, find_clumps
        >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
        >>> data_source = ds.disk([0.5, 0.5, 0.5], [0., 0., 1.],
        ...                       (8, 'kpc'), (1, 'kpc'))
        >>> field = ("gas", "density")
        >>> step = 2.0
        >>> c_min = 10**np.floor(np.log10(data_source[field]).min()  )
        >>> c_max = 10**np.floor(np.log10(data_source[field]).max()+1)
        >>> master_clump = Clump(data_source, field)
        >>> master_clump.add_info_item("center_of_mass")
        >>> master_clump.add_validator("min_cells", 20)
        >>> find_clumps(master_clump, c_min, c_max, step)
        >>> fn = master_clump.save_as_dataset(fields=["density", "particle_mass"])
        >>> new_ds = yt.load(fn)
        >>> print (ds.tree["clump", "cell_mass"])
        1296926163.91 Msun
        >>> print ds.tree["grid", "density"]
        [  2.54398434e-26   2.46620353e-26   2.25120154e-26 ...,   1.12879234e-25
           1.59561490e-25   1.09824903e-24] g/cm**3
        >>> print ds.tree["all", "particle_mass"]
        [  4.25472446e+38   4.25472446e+38   4.25472446e+38 ...,   2.04238266e+38
           2.04523901e+38   2.04770938e+38] g
        >>> print ds.tree.children[0]["clump", "cell_mass"]
        909636495.312 Msun
        >>> print ds.leaves[0]["clump", "cell_mass"]
        3756566.99809 Msun
        >>> print ds.leaves[0]["grid", "density"]
        [  6.97820274e-24   6.58117370e-24   7.32046082e-24   6.76202430e-24
           7.41184837e-24   6.76981480e-24   6.94287213e-24   6.56149658e-24
           6.76584569e-24   6.94073710e-24   7.06713082e-24   7.22556526e-24
           7.08338898e-24   6.78684331e-24   7.40647040e-24   7.03050456e-24
           7.12438678e-24   6.56310217e-24   7.23201662e-24   7.17314333e-24] g/cm**3

        """

        ds = self.data.ds
        keyword = "%s_clump_%d" % (str(ds), self.clump_id)
        filename = get_output_filename(filename, keyword, ".h5")

        # collect clump info fields
        clump_info = dict([(ci.name, []) for ci in self.base.clump_info])
        clump_info.update(
            dict([(field, []) for field in ["clump_id", "parent_id",
                                            "contour_key", "contour_id"]]))
        for clump in self:
            clump_info["clump_id"].append(clump.clump_id)
            if clump.parent is None:
                parent_id = -1
            else:
                parent_id = clump.parent.clump_id
            clump_info["parent_id"].append(parent_id)

            contour_key = clump.contour_key
            if contour_key is None: contour_key = -1
            clump_info["contour_key"].append(contour_key)
            contour_id = clump.contour_id
            if contour_id is None: contour_id = -1
            clump_info["contour_id"].append(contour_id)

            for ci in self.base.clump_info:
                clump_info[ci.name].append(clump.info[ci.name][1])
        for ci in clump_info:
            if hasattr(clump_info[ci][0], "units"):
                clump_info[ci] = ds.arr(clump_info[ci])
            else:
                clump_info[ci] = np.array(clump_info[ci])

        ftypes = dict([(ci, "clump") for ci in clump_info])

        # collect data fields
        if fields is not None:
            contour_fields = \
              [("index", "contours_%s" % ckey)
               for ckey in np.unique(clump_info["contour_key"]) \
               if str(ckey) != "-1"]

            ptypes = []
            field_data = {}
            need_grid_positions = False
            for f in self.base.data._determine_fields(fields) + contour_fields:
                if ds.field_info[f].particle_type:
                    if f[0] not in ptypes:
                        ptypes.append(f[0])
                    ftypes[f] = f[0]
                else:
                    need_grid_positions = True
                    if f[1] in ('x', 'y', 'z', 'dx', 'dy', 'dz'):
                        # skip 'xyz' if a user passes that in because they
                        # will be added to ftypes below
                        continue
                    ftypes[f] = "grid"
                field_data[f] = self.base[f]

            if len(ptypes) > 0:
                for ax in "xyz":
                    for ptype in ptypes:
                        p_field = (ptype, "particle_position_%s" % ax)
                        if p_field in ds.field_info and \
                          p_field not in field_data:
                            ftypes[p_field] = p_field[0]
                            field_data[p_field] = self.base[p_field]

                for clump in self:
                    if clump.contour_key is None:
                        continue
                    for ptype in ptypes:
                        cfield = (ptype, "contours_%s" % clump.contour_key)
                        if cfield not in field_data:
                            field_data[cfield] = \
                              clump.data._part_ind(ptype).astype(np.int64)
                            ftypes[cfield] = ptype
                        field_data[cfield][clump.data._part_ind(ptype)] = \
                          clump.contour_id

            if need_grid_positions:
                for ax in "xyz":
                    g_field = ("index", ax)
                    if g_field in ds.field_info and \
                      g_field not in field_data:
                        field_data[g_field] = self.base[g_field]
                        ftypes[g_field] = "grid"
                    g_field = ("index", "d" + ax)
                    if g_field in ds.field_info and \
                      g_field not in field_data:
                        ftypes[g_field] = "grid"
                        field_data[g_field] = self.base[g_field]

            if self.contour_key is not None:
                cfilters = {}
                for field in field_data:
                    if ftypes[field] == "grid":
                        ftype = "index"
                    else:
                        ftype = field[0]
                    cfield = (ftype, "contours_%s" % self.contour_key)
                    if cfield not in cfilters:
                        cfilters[cfield] = field_data[cfield] == self.contour_id
                    field_data[field] = field_data[field][cfilters[cfield]]

            clump_info.update(field_data)
        extra_attrs = {"data_type": "yt_clump_tree",
                       "container_type": "yt_clump_tree"}
        save_as_dataset(ds, filename, clump_info,
                        field_types=ftypes,
                        extra_attrs=extra_attrs)

        return filename

    def pass_down(self,operation):
        """
        Performs an operation on a clump with an exec and passes the 
        instruction down to clump children.
        """

        # Call if callable, otherwise do an exec.
        if callable(operation):
            operation()
        else:
            exec(operation)

        for child in self.children:
            child.pass_down(operation)

    def _validate(self):
        "Apply all user specified validator functions."

        # Only call functions if not done already.
        if self.valid is not None:
            return self.valid

        self.valid = True
        for validator in self.validators:
            self.valid &= validator(self)
            if not self.valid:
                break

        return self.valid

    def __reduce__(self):
        raise RuntimeError(
            "Pickling Clump instances is not supported. Please use "
            "Clump.save_as_dataset instead")
    
    def __getitem__(self, request):
        return self.data[request]

def find_clumps(clump, min_val, max_val, d_clump):
    mylog.info("Finding clumps: min: %e, max: %e, step: %f" % 
               (min_val, max_val, d_clump))
    if min_val >= max_val:
        return
    clump.find_children(min_val, max_val=max_val)

    if len(clump.children) == 1:
        find_clumps(clump, min_val*d_clump, max_val, d_clump)

    elif len(clump.children) > 0:
        these_children = []
        mylog.info("Investigating %d children." % len(clump.children))
        for child in clump.children:
            find_clumps(child, min_val*d_clump, max_val, d_clump)
            if len(child.children) > 0:
                these_children.append(child)
            elif (child._validate()):
                these_children.append(child)
            else:
                mylog.info(("Eliminating invalid, childless clump with " +
                            "%d cells.") % len(child.data["ones"]))
        if len(these_children) > 1:
            mylog.info("%d of %d children survived." %
                       (len(these_children),len(clump.children)))
            clump.children = these_children
        elif len(these_children) == 1:
            mylog.info(("%d of %d children survived, linking its " +
                        "children to parent.") % 
                        (len(these_children),len(clump.children)))
            clump.children = these_children[0].children
            for child in clump.children:
                child.parent = clump
                child.data.parent = clump.data
        else:
            mylog.info("%d of %d children survived, erasing children." %
                       (len(these_children),len(clump.children)))
            clump.children = []

def get_lowest_clumps(clump, clump_list=None):
    "Return a list of all clumps at the bottom of the index."

    issue_deprecation_warning(
        "This function has been deprecated in favor of accessing a " +
        "clump's leaf nodes via 'clump.leaves'.")
    return clump.leaves



"""
Various definitions for various other modules and routines



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------





import numpy as np

from yt.testing import \
    fake_random_ds, \
    assert_equal


def test_ortho_ray():
    ds = fake_random_ds(64, nprocs=8)
    dx = (ds.domain_right_edge - ds.domain_left_edge) / \
          ds.domain_dimensions

    axes = ['x', 'y', 'z']
    for ax, an in enumerate(axes):
        ocoord = ds.arr(np.random.random(2), 'code_length')

        my_oray = ds.ortho_ray(ax, ocoord)

        my_axes = ds.coordinates.x_axis[ax], ds.coordinates.y_axis[ax]

        # find the cells intersected by the ortho ray
        my_all = ds.all_data()
        my_cells = (np.abs(my_all[axes[my_axes[0]]] - ocoord[0]) <= 
                    0.5 * dx[my_axes[0]]) & \
                   (np.abs(my_all[axes[my_axes[1]]] - ocoord[1]) <= 
                    0.5 * dx[my_axes[1]])

        assert_equal(my_oray['density'].sum(),
                     my_all['density'][my_cells].sum())



"""
API for yt.visualization



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from .color_maps import \
    add_colormap, \
    show_colormaps, \
    make_colormap

from .particle_plots import \
    ParticleProjectionPlot, \
    ParticlePhasePlot, \
    ParticlePlot

from .fixed_resolution import \
    FixedResolutionBuffer, \
    ObliqueFixedResolutionBuffer, \
    ParticleImageBuffer

from .image_writer import \
    multi_image_composite, \
    write_bitmap, \
    write_image, \
    map_to_colors, \
    splat_points, \
    apply_colormap, \
    scale_image, \
    write_projection

from .plot_modifications import \
    PlotCallback, \
    callback_registry

from .streamlines import \
    Streamlines

from .plot_window import \
    SlicePlot, \
    AxisAlignedSlicePlot, \
    OffAxisSlicePlot, \
    ProjectionPlot, \
    OffAxisProjectionPlot, \
    plot_2d

from .line_plot import \
    LinePlot, \
    LineBuffer

from .profile_plotter import \
    ProfilePlot, \
    PhasePlot

from .base_plot_types import \
    get_multi_plot

from .fits_image import \
    FITSImageData, \
    FITSSlice, \
    FITSOffAxisSlice, \
    FITSProjection, \
    FITSOffAxisProjection



import nose

from yt.config import ytcfg
from yt.testing import fake_random_ds, requires_module

def setup_func():
    ytcfg["yt", "__withintesting"] = "True"

@requires_module('glue.core')
@nose.with_setup(setup_func)
def test_glue_data_object():
    ds = fake_random_ds(16)
    ad = ds.all_data()
    ad.to_glue([('gas', 'density')])



"""
AdaptaHOP frontend tests



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2015, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from yt.testing import requires_file
from yt.utilities.answer_testing.framework import \
    data_dir_load

from yt.frontends.adaptahop.data_structures import AdaptaHOPDataset

r0 = "output_00080/info_00080.txt"
r1 = "output_00080_halos/tree_bricks080"

@requires_file(r0)
@requires_file(r1)
def test_opening():
    ds_parent = data_dir_load(r0)
    ds = data_dir_load(r1, kwargs=dict(parent_ds=ds_parent))

    ds.index

    assert isinstance(ds, AdaptaHOPDataset)

@requires_file(r0)
@requires_file(r1)
def test_field_access():
    ds_parent = data_dir_load(r0)
    ds = data_dir_load(r1, kwargs=dict(parent_ds=ds_parent))

    fields = [(ptype, field) for (ptype, field) in ds.derived_field_list
              if (ptype == 'halos') and (field != 'particle_identities')]

    ad = ds.all_data()
    def test_access(ptype, field):
        ad[ptype, field]
    for (ptype, field) in fields:
        test_access(ptype, field)

@requires_file(r0)
@requires_file(r1)
def test_get_halo():
    ds_parent = data_dir_load(r0)
    ds = data_dir_load(r1, kwargs=dict(parent_ds=ds_parent))

    halo = ds.halo(1, ptype='io')

    # Check halo objet has position, velocity, mass and members attributes
    for attr_name in ('mass', 'position', 'velocity', 'member_ids'):
        getattr(halo, attr_name)

    members = np.sort(halo.member_ids)

    # Check sphere contains all the members
    id_sphere = halo.sphere['io', 'particle_identity'].astype(int)
    assert len(np.lib.arraysetops.setdiff1d(members, id_sphere)) == 0

    # Check region contains *only* halo particles
    id_reg = np.sort(halo['io', 'particle_identity'].astype(int))

    np.testing.assert_equal(members, id_reg)



"""
AHF data structures



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2017, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import glob
import os
import stat

import numpy as np

from yt.data_objects.static_output import \
    Dataset
from yt.frontends.halo_catalog.data_structures import \
    HaloCatalogFile
from yt.funcs import \
    setdefaultattr
from yt.geometry.particle_geometry_handler import \
    ParticleIndex
from yt.utilities.cosmology import \
    Cosmology

from .fields import AHFHalosFieldInfo


class AHFHalosFile(HaloCatalogFile):
    def __init__(self, ds, io, filename, file_id):
        root, _ = os.path.splitext(filename)
        candidates = glob.glob(root + '*.AHF_halos')
        if len(candidates) == 1:
            filename = candidates[0]
        else:
            raise ValueError('Too many AHF_halos files.')
        self.col_names = self._read_column_names(filename)
        super(AHFHalosFile, self).__init__(ds, io, filename, file_id)

    def read_data(self, usecols=None):
        return np.genfromtxt(self.filename, names=self.col_names,
                             usecols=usecols)

    def _read_column_names(self, filename):
        with open(filename) as f:
            line = f.readline()
            # Remove leading '#'
            line = line[1:]
            names = line.split()
            # Remove trailing '()'
            names = [name.split('(')[0] for name in names]
            return names

    def _read_particle_positions(self, ptype, f=None):
        """
        Read all particle positions in this file.
        """

        halos = self.read_data(usecols=['Xc', 'Yc', 'Zc'])
        pos = np.empty((halos.size, 3), dtype="float64")
        for i, ax in enumerate('XYZ'):
            pos[:, i] = halos['%sc' % ax].astype('float64')

        return pos

class AHFHalosDataset(Dataset):
    _index_class = ParticleIndex
    _file_class = AHFHalosFile
    _field_info_class = AHFHalosFieldInfo

    def __init__(self, filename, dataset_type='ahf',
                 n_ref=16, over_refine_factor=1,
                 units_override=None, unit_system='cgs',
                 hubble_constant=1.0):
        root, _ = os.path.splitext(filename)
        self.log_filename = root + '.log'
        self.hubble_constant = hubble_constant

        self.n_ref = n_ref
        self.over_refine_factor = over_refine_factor
        super(AHFHalosDataset, self).__init__(
            filename, dataset_type=dataset_type,
            units_override=units_override, unit_system=unit_system
        )

    def _set_code_unit_attributes(self):
        setdefaultattr(self, 'length_unit', self.quan(1.0, 'kpccm/h'))
        setdefaultattr(self, 'mass_unit', self.quan(1.0, 'Msun/h'))
        setdefaultattr(self, 'time_unit', self.quan(1.0, 's'))
        setdefaultattr(self, 'velocity_unit', self.quan(1.0, 'km/s'))

    def _parse_parameter_file(self):
        # Read all parameters.
        simu = self._read_log_simu()
        param = self._read_parameter()

        # Set up general information.
        self.filename_template = self.parameter_filename
        self.file_count = 1
        self.parameters.update(param)
        self.particle_types = ('halos')
        self.particle_types_raw = ('halos')
        self.unique_identifier = \
            int(os.stat(self.parameter_filename)[stat.ST_CTIME])

        # Set up geometrical information.
        self.refine_by = 2
        self.dimensionality = 3
        nz = 1 << self.over_refine_factor
        self.domain_dimensions = np.ones(self.dimensionality, "int32") * nz
        self.domain_left_edge = np.array([0.0, 0.0, 0.0])
        # Note that boxsize is in Mpc but particle positions are in kpc.
        self.domain_right_edge = np.array([simu['boxsize']] * 3) * 1000
        self.periodicity = (True, True, True)

        # Set up cosmological information.
        self.cosmological_simulation = 1
        self.current_redshift = param['z']
        self.omega_lambda = simu['lambda0']
        self.omega_matter = simu['omega0']
        cosmo = Cosmology(hubble_constant=self.hubble_constant,
                          omega_matter=self.omega_matter,
                          omega_lambda=self.omega_lambda)
        self.current_time = cosmo.lookback_time(param['z'], 1e6).in_units('s')

    @classmethod
    def _is_valid(self, *args, **kwargs):
        filename = args[0]
        if not filename.endswith('.parameter'):
            return False
        with open(filename, 'r') as f:
            if f.readlines()[11].startswith('AHF'):
                return True
        return False

    # Helper methods

    def _read_log_simu(self):
        simu = {}
        with open(self.log_filename) as f:
            for l in f:
                if l.startswith('simu.'):
                    name, val = l.split(':')
                    key = name.strip().split('.')[1]
                    try:
                        val = float(val)
                    except Exception:
                        val = float.fromhex(val)
                    simu[key] = val
        return simu

    def _read_parameter(self):
        param = {}
        with open(self.parameter_filename) as f:
            for l in f:
                words = l.split()
                if len(words) == 2:
                    key, val = words
                    try:
                        val = float(val)
                        param[key] = val
                    except Exception: pass
        return param

    @property
    def _skip_cache(self):
        return True



'''
AHF-specific fields



'''

#-----------------------------------------------------------------------------
# Copyright (c) 2017, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.fields.field_info_container import \
    FieldInfoContainer

m_units = 'Msun/h'
p_units = 'kpccm/h'
r_units = 'kpccm/h'
v_units = 'km/s'


class AHFHalosFieldInfo(FieldInfoContainer):
    known_other_fields = ()

    # See http://popia.ft.uam.es/AHF/files/AHF.pdf
    # and search for '*.AHF_halos'.
    known_particle_fields = (
        ('ID', ('', ['particle_identifier'], None)),
        ('hostHalo', ('', [], None)),
        ('numSubStruct', ('', [], None)),
        ('Mvir', (m_units, ['particle_mass'], 'Virial Mass')),
        ('npart', ('', [], None)),
        ('Xc', (p_units, ['particle_position_x'], None)),
        ('Yc', (p_units, ['particle_position_y'], None)),
        ('Zc', (p_units, ['particle_position_z'], None)),
        ('VXc', (v_units, ['particle_velocity_x'], None)),
        ('VYc', (v_units, ['particle_velocity_y'], None)),
        ('VZc', (v_units, ['particle_velocity_z'], None)),
        ('Rvir', (r_units, ['virial_radius'], 'Virial Radius')),
        ('Rmax', (r_units, [], None)),
        ('r2', (r_units, [], None)),
        ('mbp_offset', (r_units, [], None)),
        ('com_offset', (r_units, [], None)),
        ('Vmax', (v_units, [], None)),
        ('v_sec', (v_units, [], None)),
        ('sigV', (v_units, [], None)),
        ('lambda', ('', [], None)),
        ('lambdaE', ('', [], None)),
        ('Lx', ('', [], None)),
        ('Ly', ('', [], None)),
        ('Lz', ('', [], None)),
        ('b', ('', [], None)),
        ('c', ('', [], None)),
        ('Eax', ('', [], None)),
        ('Eay', ('', [], None)),
        ('Eaz', ('', [], None)),
        ('Ebx', ('', [], None)),
        ('Eby', ('', [], None)),
        ('Ebz', ('', [], None)),
        ('Ecx', ('', [], None)),
        ('Ecy', ('', [], None)),
        ('Ecz', ('', [], None)),
        ('ovdens', ('', [], None)),
        ('nbins', ('', [], None)),
        ('fMhires', ('', [], None)),
        ('Ekin', ('Msun/h*(km/s)**2', [], None)),
        ('Epot', ('Msun/h*(km/s)**2', [], None)),
        ('SurfP', ('Msun/h*(km/s)**2', [], None)),
        ('Phi0', ('(km/s)**2', [], None)),
        ('cNFW', ('', [], None))
    )



"""
Tipsy data-file handling function




"""
from __future__ import print_function

#-----------------------------------------------------------------------------
# Copyright (c) 2014, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import glob
import numpy as np
from numpy.lib.recfunctions import append_fields
import os

from yt.utilities.io_handler import \
    BaseIOHandler
from yt.utilities.lib.geometry_utils import \
    compute_morton
from yt.utilities.logger import ytLogger as \
    mylog

CHUNKSIZE = 10000000


class IOHandlerTipsyBinary(BaseIOHandler):
    _dataset_type = "tipsy"
    _vector_fields = ("Coordinates", "Velocity", "Velocities")

    _pdtypes = None  # dtypes, to be filled in later
    _aux_pdtypes = None  # auxiliary files' dtypes

    _ptypes = ("Gas",
               "DarkMatter",
               "Stars")
    _chunksize = 64 * 64 * 64

    _aux_fields = None
    _fields = (("Gas", "Mass"),
               ("Gas", "Coordinates"),
               ("Gas", "Velocities"),
               ("Gas", "Density"),
               ("Gas", "Temperature"),
               ("Gas", "Epsilon"),
               ("Gas", "Metals"),
               ("Gas", "Phi"),
               ("DarkMatter", "Mass"),
               ("DarkMatter", "Coordinates"),
               ("DarkMatter", "Velocities"),
               ("DarkMatter", "Epsilon"),
               ("DarkMatter", "Phi"),
               ("Stars", "Mass"),
               ("Stars", "Coordinates"),
               ("Stars", "Velocities"),
               ("Stars", "Metals"),
               ("Stars", "FormationTime"),
               ("Stars", "Epsilon"),
               ("Stars", "Phi")
               )

    def __init__(self, *args, **kwargs):
        self._aux_fields = []
        super(IOHandlerTipsyBinary, self).__init__(*args, **kwargs)

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError

    def _fill_fields(self, fields, vals, mask, data_file):
        if mask is None:
            size = 0
        else:
            size = mask.sum()
        rv = {}
        for field in fields:
            mylog.debug("Allocating %s values for %s", size, field)
            if field in self._vector_fields:
                rv[field] = np.empty((size, 3), dtype="float64")
                if size == 0:
                    continue
                rv[field][:, 0] = vals[field]['x'][mask]
                rv[field][:, 1] = vals[field]['y'][mask]
                rv[field][:, 2] = vals[field]['z'][mask]
            else:
                rv[field] = np.empty(size, dtype="float64")
                if size == 0:
                    continue
                rv[field][:] = vals[field][mask]
            if field == "Coordinates":
                eps = np.finfo(rv[field].dtype).eps
                for i in range(3):
                    rv[field][:, i] = np.clip(rv[field][:, i],
                                              self.domain_left_edge[i] + eps,
                                              self.domain_right_edge[i] - eps)
        return rv

    def _read_particle_coords(self, chunks, ptf):
        data_files = set([])
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        for data_file in sorted(data_files):
            poff = data_file.field_offsets
            tp = data_file.total_particles
            f = open(data_file.filename, "rb")
            for ptype, field_list in sorted(ptf.items(),
                                            key=lambda a: poff[a[0]]):
                f.seek(poff[ptype], os.SEEK_SET)
                total = 0
                while total < tp[ptype]:
                    count = min(self._chunksize, tp[ptype] - total)
                    p = np.fromfile(f, self._pdtypes[ptype], count=count)
                    total += p.size
                    d = [p["Coordinates"][ax].astype("float64")
                         for ax in 'xyz']
                    del p
                    yield ptype, d

    def _read_particle_fields(self, chunks, ptf, selector):
        chunks = list(chunks)
        data_files = set([])
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        for data_file in sorted(data_files):
            poff = data_file.field_offsets
            aux_fields_offsets = \
                self._calculate_particle_offsets_aux(data_file)
            tp = data_file.total_particles
            f = open(data_file.filename, "rb")

            # we need to open all aux files for chunking to work
            aux_fh = {}
            for afield in self._aux_fields:
                aux_fh[afield] = open(data_file.filename + '.' + afield, 'rb')

            for ptype, field_list in sorted(ptf.items(),
                                            key=lambda a: poff[a[0]]):
                f.seek(poff[ptype], os.SEEK_SET)
                afields = list(set(field_list).intersection(self._aux_fields))
                for afield in afields:
                    aux_fh[afield].seek(
                        aux_fields_offsets[afield][ptype][0], os.SEEK_SET)

                total = 0
                while total < tp[ptype]:
                    count = min(self._chunksize, tp[ptype] - total)
                    p = np.fromfile(f, self._pdtypes[ptype], count=count)

                    auxdata = []
                    for afield in afields:
                        if isinstance(self._aux_pdtypes[afield], np.dtype):
                            auxdata.append(
                                np.fromfile(aux_fh[afield],
                                            self._aux_pdtypes[afield],
                                            count=count)
                            )
                        else:
                            aux_fh[afield].seek(0, os.SEEK_SET)
                            sh = aux_fields_offsets[afield][ptype][0] + total
                            sf = aux_fields_offsets[afield][ptype][1] + \
                                tp[ptype] - count - total
                            if tp[ptype] > 0:
                                aux = np.genfromtxt(
                                    aux_fh[afield], skip_header=sh,
                                    skip_footer=sf
                                )
                                if aux.ndim < 1:
                                    aux = np.array([aux])
                                auxdata.append(aux)

                    total += p.size
                    if afields:
                        p = append_fields(p, afields, auxdata)
                    mask = selector.select_points(
                        p["Coordinates"]['x'].astype("float64"),
                        p["Coordinates"]['y'].astype("float64"),
                        p["Coordinates"]['z'].astype("float64"), 0.0)
                    if mask is None:
                        continue
                    tf = self._fill_fields(field_list, p, mask, data_file)
                    for field in field_list:
                        yield (ptype, field), tf.pop(field)

            # close all file handles
            f.close()
            for fh in list(aux_fh.values()):
                fh.close()

    def _update_domain(self, data_file):
        '''
        This method is used to determine the size needed for a box that will
        bound the particles.  It simply finds the largest position of the
        whole set of particles, and sets the domain to +/- that value.
        '''
        ds = data_file.ds
        ind = 0
        # Check to make sure that the domain hasn't already been set
        # by the parameter file
        if np.all(np.isfinite(ds.domain_left_edge)) and \
                np.all(np.isfinite(ds.domain_right_edge)):
            return
        with open(data_file.filename, "rb") as f:
            ds.domain_left_edge = 0
            ds.domain_right_edge = 0
            f.seek(ds._header_offset)
            mi = np.array([1e30, 1e30, 1e30], dtype="float64")
            ma = -np.array([1e30, 1e30, 1e30], dtype="float64")
            for iptype, ptype in enumerate(self._ptypes):
                # We'll just add the individual types separately
                count = data_file.total_particles[ptype]
                if count == 0:
                    continue
                stop = ind + count
                while ind < stop:
                    c = min(CHUNKSIZE, stop - ind)
                    pp = np.fromfile(f, dtype=self._pdtypes[ptype],
                                     count=c)
                    np.minimum(mi, [pp["Coordinates"]["x"].min(),
                                    pp["Coordinates"]["y"].min(),
                                    pp["Coordinates"]["z"].min()], mi)
                    np.maximum(ma, [pp["Coordinates"]["x"].max(),
                                    pp["Coordinates"]["y"].max(),
                                    pp["Coordinates"]["z"].max()], ma)
                    ind += c
        # We extend by 1%.
        DW = ma - mi
        mi -= 0.01 * DW
        ma += 0.01 * DW
        ds.domain_left_edge = ds.arr(mi, 'code_length')
        ds.domain_right_edge = ds.arr(ma, 'code_length')
        ds.domain_width = DW = ds.domain_right_edge - ds.domain_left_edge
        ds.unit_registry.add("unitary", float(DW.max() * DW.units.base_value),
                             DW.units.dimensions)

    def _initialize_index(self, data_file, regions):
        ds = data_file.ds
        morton = np.empty(sum(list(data_file.total_particles.values())),
                          dtype="uint64")
        ind = 0
        DLE, DRE = ds.domain_left_edge, ds.domain_right_edge
        self.domain_left_edge = DLE.in_units("code_length").ndarray_view()
        self.domain_right_edge = DRE.in_units("code_length").ndarray_view()
        with open(data_file.filename, "rb") as f:
            f.seek(ds._header_offset)
            for iptype, ptype in enumerate(self._ptypes):
                # We'll just add the individual types separately
                count = data_file.total_particles[ptype]
                if count == 0:
                    continue
                stop = ind + count
                while ind < stop:
                    c = min(CHUNKSIZE, stop - ind)
                    pp = np.fromfile(f, dtype=self._pdtypes[ptype],
                                     count=c)
                    mis = np.empty(3, dtype="float64")
                    mas = np.empty(3, dtype="float64")
                    for axi, ax in enumerate('xyz'):
                        mi = pp["Coordinates"][ax].min()
                        ma = pp["Coordinates"][ax].max()
                        mylog.debug(
                            "Spanning: %0.3e .. %0.3e in %s", mi, ma, ax)
                        mis[axi] = mi
                        mas[axi] = ma
                    pos = np.empty((pp.size, 3), dtype="float64")
                    for i, ax in enumerate("xyz"):
                        pos[:, i] = pp["Coordinates"][ax]
                    regions.add_data_file(pos, data_file.file_id,
                                          data_file.ds.filter_bbox)
                    morton[ind:ind + c] = compute_morton(
                        pos[:, 0], pos[:, 1], pos[:, 2],
                        DLE, DRE, data_file.ds.filter_bbox)
                    ind += c
        mylog.info("Adding %0.3e particles", morton.size)
        return morton

    def _count_particles(self, data_file):
        npart = {
            "Gas": data_file.ds.parameters['nsph'],
            "Stars": data_file.ds.parameters['nstar'],
            "DarkMatter": data_file.ds.parameters['ndark']
        }
        return npart

    @classmethod
    def _compute_dtypes(cls, field_dtypes, endian="<"):
        pds = {}
        for ptype, field in cls._fields:
            dtbase = field_dtypes.get(field, 'f')
            ff = "%s%s" % (endian, dtbase)
            if field in cls._vector_fields:
                dt = (field, [('x', ff), ('y', ff), ('z', ff)])
            else:
                dt = (field, ff)
            pds.setdefault(ptype, []).append(dt)
        pdtypes = {}
        for ptype in pds:
            pdtypes[ptype] = np.dtype(pds[ptype])
        return pdtypes

    def _create_dtypes(self, data_file):
        # We can just look at the particle counts.
        self._header_offset = data_file.ds._header_offset
        self._pdtypes = self._compute_dtypes(data_file.ds._field_dtypes,
                                             data_file.ds.endian)
        self._field_list = []
        for ptype, field in self._fields:
            if data_file.total_particles[ptype] == 0:
                # We do not want out _pdtypes to have empty particles.
                self._pdtypes.pop(ptype, None)
                continue
            self._field_list.append((ptype, field))

        # Find out which auxiliaries we have and what is their format
        tot_parts = np.sum(list(data_file.total_particles.values()))
        endian = data_file.ds.endian
        self._aux_pdtypes = {}
        self._aux_fields = []
        for f in glob.glob(data_file.filename + '.*'):
            afield = f.rsplit('.')[-1]
            filename = data_file.filename + '.' + afield
            if not os.path.exists(filename):
                continue
            self._aux_fields.append(afield)
        skip_afields = []
        for afield in self._aux_fields:
            filename = data_file.filename + '.' + afield
            # We need to do some fairly ugly detection to see what format the
            # auxiliary files are in.  They can be either ascii or binary, and
            # the binary files can be either floats, ints, or doubles.  We're
            # going to use a try-catch cascade to determine the format.
            filesize = os.stat(filename).st_size
            if np.fromfile(filename, np.dtype(endian + 'i4'),
                           count=1) != tot_parts:
                with open(filename, 'rb') as f:
                    header_nparts = f.readline()
                    try:
                        header_nparts = int(header_nparts)
                    except ValueError:
                        skip_afields.append(afield)
                        continue
                    if int(header_nparts) != tot_parts:
                        raise RuntimeError
                self._aux_pdtypes[afield] = "ascii"
            elif (filesize - 4) / 8 == tot_parts:
                self._aux_pdtypes[afield] = np.dtype([('aux', endian + 'd')])
            elif (filesize - 4) / 4 == tot_parts:
                if afield.startswith("i"):
                    self._aux_pdtypes[afield] = np.dtype([('aux', endian + 'i')])
                else:
                    self._aux_pdtypes[afield] = np.dtype([('aux', endian + 'f')])
            else:
                skip_afields.append(afield)
        for afield in skip_afields:
            self._aux_fields.remove(afield)
        # Add the auxiliary fields to each ptype we have
        for ptype in self._ptypes:
            if any([ptype == field[0] for field in self._field_list]):
                self._field_list += \
                    [(ptype, afield) for afield in self._aux_fields]
        return self._field_list

    def _identify_fields(self, data_file):
        return self._field_list, {}

    def _calculate_particle_offsets(self, data_file):
        field_offsets = {}
        pos = data_file.ds._header_offset
        for ptype in self._ptypes:
            field_offsets[ptype] = pos
            if data_file.total_particles[ptype] == 0:
                continue
            size = self._pdtypes[ptype].itemsize
            pos += data_file.total_particles[ptype] * size
        return field_offsets

    def _calculate_particle_offsets_aux(self, data_file):
        aux_fields_offsets = {}
        tp = data_file.total_particles
        for afield in self._aux_fields:
            aux_fields_offsets[afield] = {}
            if isinstance(self._aux_pdtypes[afield], np.dtype):
                pos = 4  # i4
                for ptype in self._ptypes:
                    aux_fields_offsets[afield][ptype] = (pos, 0)
                    if data_file.total_particles[ptype] == 0:
                        continue
                    size = np.dtype(self._aux_pdtypes[afield]).itemsize
                    pos += data_file.total_particles[ptype] * size
            else:
                aux_fields_offsets[afield].update(
                    {'Gas': (1, tp["DarkMatter"] + tp["Stars"]),
                     'DarkMatter': (1 + tp["Gas"], tp["Stars"]),
                     'Stars': (1 + tp["DarkMatter"] + tp["Gas"], 0)}
                )
        return aux_fields_offsets



"""
Data structures for OWLS frontend




"""
from __future__ import print_function

#-----------------------------------------------------------------------------
# Copyright (c) 2014, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.utilities.on_demand_imports import _h5py as h5py

import yt.units
from yt.frontends.gadget.data_structures import \
    GadgetHDF5Dataset
from yt.utilities.definitions import \
    sec_conversion

from .fields import \
    OWLSFieldInfo

class OWLSDataset(GadgetHDF5Dataset):
    _particle_mass_name = "Mass"
    _field_info_class = OWLSFieldInfo
    _time_readin = "Time_GYR"


    def _parse_parameter_file(self):

        # read values from header
        hvals = self._get_hvals()
        self.parameters = hvals

        # set features common to OWLS and Eagle
        self._set_owls_eagle()

        # Set time from value in header
        self.current_time = hvals[self._time_readin] * \
                            sec_conversion["Gyr"] * yt.units.s


    def _set_code_unit_attributes(self):
        self._set_owls_eagle_units()


    @classmethod
    def _is_valid(self, *args, **kwargs):
        need_groups = ['Constants', 'Header', 'Parameters', 'Units']
        veto_groups = ['SUBFIND', 'FOF',
                       'PartType0/ChemistryAbundances', 
                       'PartType0/ChemicalAbundances',
                       'RuntimePars', 'HashTable']
        valid = True
        try:
            fileh = h5py.File(args[0], mode='r')
            for ng in need_groups:
                if ng not in fileh["/"]:
                    valid = False
            for vg in veto_groups:
                if vg in fileh["/"]:
                    valid = False                    
            fileh.close()
        except Exception:
            valid = False
            pass
        return valid



"""
Title: answer_tests.py
Purpose: Contains answer tests that are used by yt's various frontends
"""
import os
import tempfile

import hashlib
import matplotlib.image as mpimg
import numpy as np

from yt.analysis_modules.cosmological_observation.api import \
     LightCone
from yt.analysis_modules.halo_analysis.api import HaloCatalog
from yt.analysis_modules.halo_mass_function.api import HaloMassFcn
from yt.utilities.on_demand_imports import \
    _h5py as h5py
from . import utils
import yt.visualization.plot_window as pw


def grid_hierarchy(ds):
    result = {}
    result['grid_dimensions'] = ds.index.grid_dimensions
    result['grid_left_edge'] = ds.index.grid_left_edge
    result['grid_right_edge'] = ds.index.grid_right_edge
    result['grid_levels'] = ds.index.grid_levels
    result['grid_particle_count'] = ds.index.grid_particle_count
    return result

def parentage_relationships(ds):
    parents = []
    children = []
    for g in ds.index.grids:
        p = g.Parent
        if p is None:
            parents.append(-1)
        elif hasattr(p, "id"):
            parents.append(p.id)
        else:
            parents = parents + [pg.id for pg in p]
        children = children + [c.id for c in g.Children]
    result = np.array(parents + children)
    return result 

def grid_values(ds, field):
    # The hashing is done here so that there is only one entry for
    # the test that contains info about all of the grids as opposed
    # to having a separate 'grid_id : grid_hash' pair for each grid
    # since that makes the answer file much larger
    result = None
    for g in ds.index.grids:
        if result is None:
            result = hashlib.md5(bytes(g.id) + g[field].tobytes())
        else:
            result.update(bytes(g.id) + g[field].tobytes())
        g.clear_data()
    return result.hexdigest()

def projection_values(ds, axis, field, weight_field, dobj_type):
    if dobj_type is not None:
        dobj = utils.create_obj(ds, dobj_type)
    else:
        dobj = None
    if ds.domain_dimensions[axis] == 1:
        # This originally returned None, but None can't be converted
        # to a bytes array (for hashing), so use -1 as a string,
        # since ints can't be converted to bytes either
        return bytes(str(-1).encode('utf-8'))
    proj = ds.proj(field,
                axis,
                weight_field=weight_field,
                data_source=dobj
            )
    # This is to try and remove python-specific anchors in the yaml
    # answer file. Also, using __repr__() results in weird strings
    # of strings that make comparison fail even though the data is
    # the same
    result = None 
    for k, v in proj.field_data.items():
        k = k.__repr__().encode('utf8')
        if result is None:
            result = hashlib.md5(k + v.tobytes())
        else:
            result.update(k + v.tobytes())
    return result.hexdigest()

def field_values(ds, field, obj_type=None, particle_type=False):
    # If needed build an instance of the dataset type
    obj = utils.create_obj(ds, obj_type)
    determined_field = obj._determine_fields(field)[0]
    # Get the proper weight field depending on if we're looking at
    # particles or not
    if particle_type:
        weight_field = (determined_field[0], "particle_ones")
    else:
        weight_field = ("index", "ones")
    # Get the average, min, and max
    avg = obj.quantities.weighted_average_quantity(
        determined_field,
        weight=weight_field)
    minimum, maximum = obj.quantities.extrema(field)
    # Return as a hashable bytestring
    return np.array([avg, minimum, maximum])

def pixelized_projection_values(ds, axis, field,
    weight_field=None, dobj_type=None):
    if dobj_type is not None:
        obj = utils.create_obj(ds, dobj_type)
    else:
        obj = None
    proj = ds.proj(field, axis, weight_field=weight_field, data_source=obj)
    frb = proj.to_frb((1.0, 'unitary'), 256)
    frb[field]
    if weight_field is not None:
        frb[weight_field]
    d = frb.data
    for f in proj.field_data:
        # Sometimes f will be a tuple.
        d["%s_sum" % (f,)] = proj.field_data[f].sum(dtype="float64")
    # This is to try and remove python-specific anchors in the yaml
    # answer file. Also, using __repr__() results in weird strings
    # of strings that make comparison fail even though the data is
    # the same
    result = None 
    for k, v in d.items():
        k = k.__repr__().encode('utf8')
        if result is None:
            result = hashlib.md5(k + v.tobytes())
        else:
            result.update(k + v.tobytes())
    return result.hexdigest()


def simulated_halo_mass_function(ds, finder):
    hc = HaloCatalog(data_ds=ds, finder_method=finder)
    hc.create()
    hmf = HaloMassFcn(halos_ds=hc.halos_ds)
    result = np.empty((2, hmf.masses_sim.size))
    result[0] = hmf.masses_sim.d
    result[1] = hmf.n_cumulative_sim.d
    return result

def analytic_halo_mass_function(ds, fit):
    hmf = HaloMassFcn(simulation_ds=ds, fitting_function=fit)
    result = np.empty((2, hmf.masses_analytic.size))
    result[0] = hmf.masses_analytic.d
    result[1] = hmf.n_cumulative_analytic.d
    return result

def small_patch_amr(ds, field, weight, axis, ds_obj):
    hex_digests = {} 
    # Grid hierarchy test
    gh_hd = grid_hierarchy(ds)
    hex_digests['grid_hierarchy'] = gh_hd
    # Parentage relationships test
    pr_hd = parentage_relationships(ds)
    hex_digests['parentage_relationships'] = pr_hd
    # Grid values, projection values, and field values tests
    gv_hd = grid_values(ds, field)
    hex_digests['grid_values'] = gv_hd
    fv_hd = field_values(ds, field, ds_obj)
    hex_digests['field_values'] = fv_hd
    pv_hd = projection_values(ds, axis, field, weight, ds_obj)
    hex_digests['projection_values'] = pv_hd
    return hex_digests

def big_patch_amr(ds, field, weight, axis, ds_obj):
    hex_digests = {} 
    # Grid hierarchy test
    gh_hd = grid_hierarchy(ds)
    hex_digests['grid_hierarchy'] = gh_hd
    # Parentage relationships test
    pr_hd = parentage_relationships(ds)
    hex_digests['parentage_relationships'] = pr_hd
    # Grid values, projection values, and field values tests
    gv_hd = grid_values(ds, field)
    hex_digests['grid_values'] = gv_hd
    ppv_hd = pixelized_projection_values(ds, axis, field, weight, ds_obj)
    hex_digests['pixelized_projection_values'] = ppv_hd 
    return hex_digests

def generic_array(func, args=None, kwargs=None):
    if args is None:
        args = []
    if kwargs is None:
        kwargs = {}
    return func(*args, **kwargs)

def sph_answer(ds, ds_str_repr, ds_nparticles, field, weight, ds_obj, axis):
    # Make sure we're dealing with the right dataset
    assert str(ds) == ds_str_repr
    # Set up keys of test names
    hex_digests = {} 
    dd = ds.all_data()
    assert dd["particle_position"].shape == (ds_nparticles, 3)
    tot = sum(dd[ptype, "particle_position"].shape[0]
              for ptype in ds.particle_types if ptype != "all")
    # Check
    assert tot == ds_nparticles
    dobj = utils.create_obj(ds, ds_obj)
    s1 = dobj["ones"].sum()
    s2 = sum(mask.sum() for block, mask in dobj.blocks)
    assert s1 == s2
    if field[0] in ds.particle_types:
        particle_type = True
    else:
        particle_type = False
    if particle_type is False:
        ppv_hd = pixelized_projection_values(ds, axis, field, weight, ds_obj)
        hex_digests['pixelized_projection_values'] = ppv_hd
    fv_hd = field_values(ds, field, ds_obj, particle_type=particle_type)
    hex_digests['field_values'] = fv_hd
    return hex_digests

def get_field_size_and_mean(ds, field, geometric):
    if geometric:
        obj = ds.all_data()
    else:
        obj = ds.data
    return np.array([obj[field].size, obj[field].mean()])

def plot_window_attribute(ds, plot_field, plot_axis, attr_name,
    attr_args, plot_type='SlicePlot', callback_id='', callback_runners=[]):
    plot = utils._create_plot_window_attribute_plot(ds, plot_type, plot_field, plot_axis, {})
    for r in callback_runners:
        r(plot_field, plot)
    attr = getattr(plot, attr_name)
    attr(*attr_args[0], **attr_args[1])
    tmpfd, tmpname = tempfile.mkstemp(suffix='.png')
    os.close(tmpfd)
    plot.save(name=tmpname)
    image = mpimg.imread(tmpname)
    os.remove(tmpname)
    return image

def phase_plot_attribute(ds_fn, x_field, y_field, z_field,
             attr_name, attr_args, plot_type='PhasePlot',
             plot_kwargs={}):
    data_source = ds_fn.all_data()
    plot = utils._create_phase_plot_attribute_plot(data_source, x_field, y_field,
                            z_field, plot_type, plot_kwargs)
    attr = getattr(plot, attr_name)
    attr(*attr_args[0], **attr_args[1])
    tmpfd, tmpname = tempfile.mkstemp(suffix='.png')
    os.close(tmpfd)
    plot.save(name=tmpname)
    image = mpimg.imread(tmpname)
    os.remove(tmpname)
    return image

def generic_image(img_fname):
    img_data = mpimg.imread(img_fname)
    return img_data

def axial_pixelization(ds):
    r"""
    This test is typically used once per geometry or coordinates type.
    Feed it a dataset, and it checks that the results of basic pixelization
    don't change.
    """
    for i, axis in enumerate(ds.coordinates.axis_order):
        (bounds, center, display_center) = \
                pw.get_window_parameters(axis, ds.domain_center, None, ds)
        slc = ds.slice(axis, center[i])
        xax = ds.coordinates.axis_name[ds.coordinates.x_axis[axis]]
        yax = ds.coordinates.axis_name[ds.coordinates.y_axis[axis]]
        pix_x = ds.coordinates.pixelize(axis, slc, xax, bounds, (512, 512))
        pix_y = ds.coordinates.pixelize(axis, slc, yax, bounds, (512, 512))
        # Wipe out all NaNs
        pix_x[np.isnan(pix_x)] = 0.0
        pix_y[np.isnan(pix_y)] = 0.0
        pix_x
        pix_y
    return pix_x, pix_y 

def light_cone_projection(parameter_file, simulation_type):
    lc = LightCone(
        parameter_file, simulation_type, 0., 0.1,
        observer_redshift=0.0, time_data=False)
    lc.calculate_light_cone_solution(
        seed=123456789, filename="LC/solution.txt")
    lc.project_light_cone(
        (600.0, "arcmin"), (60.0, "arcsec"), "density",
        weight_field=None, save_stack=True)
    fh = h5py.File("LC/LightCone.h5")
    data = fh["density_None"].value
    units = fh["density_None"].attrs["units"]
    assert units == "g/cm**2"
    fh.close()
    mean = data.mean()
    mi = data[data.nonzero()].min()
    ma = data.max()
    return np.array([mean, mi, ma])

def extract_connected_sets(ds_fn, data_source, field, num_levels, min_val, max_val):
    n, all_sets = data_source.extract_connected_sets(
        field, num_levels, min_val, max_val)
    result = []
    for level in all_sets:
        for set_id in all_sets[level]:
            result.append([all_sets[level][set_id]["cell_mass"].size,
                           all_sets[level][set_id]["cell_mass"].sum()])
    result = np.array(result)
    return result

def VR_image_comparison(scene):
    tmpfd, tmpname = tempfile.mkstemp(suffix='.png')
    os.close(tmpfd)
    scene.render()
    scene.save(tmpname, sigma_clip=1.0)
    image = mpimg.imread(tmpname)
    os.remove(tmpname)
    return image



from yt.utilities.exceptions import \
    YTBoundsDefinitionError

from yt.convenience import \
    load
from yt.testing import \
    fake_random_ds, \
    requires_file
from numpy.testing import \
    assert_raises, assert_array_less, assert_allclose
import yt


def test_cic_deposit():
    ds = fake_random_ds(64, nprocs = 8, particles=64**3)
    my_reg = ds.arbitrary_grid(ds.domain_left_edge, ds.domain_right_edge,
            dims=[1, 800, 800])
    f = ("deposit", "all_cic")
    assert_raises(YTBoundsDefinitionError, my_reg.__getitem__, f)

RAMSES = 'output_00080/info_00080.txt'
RAMSES_small = 'ramses_new_format/output_00002/info_00002.txt'
ISOGAL = 'IsolatedGalaxy/galaxy0030/galaxy0030'

@requires_file(RAMSES)
def test_one_zone_octree_deposit():
    ds = load(RAMSES)

    # Get a sphere centred on the main halo
    hpos = ds.arr([0.5215110772898429, 0.5215110772898429, 0.5215110772898429], 'code_length')
    hrvir = ds.quan(0.042307235300540924, 'Mpc')

    sp = ds.sphere(hpos, hrvir*10)
    assert sp['deposit', 'io_cic'].shape == (1,)

@requires_file(RAMSES)
@requires_file(ISOGAL)
def test_mesh_sampling():
    for fn in [RAMSES, ISOGAL]:
        ds = yt.load(fn)
        ds.add_mesh_sampling_particle_field(('index', 'x'), ptype='all')
        ds.add_mesh_sampling_particle_field(('index', 'dx'), ptype='all')

        dx = ds.r['all', 'cell_index_dx']
        xc = ds.r['all', 'cell_index_x']
        xp = ds.r['all', 'particle_position_x']

        dist = xp - xc

        assert_array_less(dist, dx)
        assert_array_less(-dist, dx)

@requires_file(RAMSES)
@requires_file(ISOGAL)
def test_mesh_sampling_for_filtered_particles():
    for fn in [RAMSES, ISOGAL]:
        ds = yt.load(RAMSES)

        @yt.particle_filter(requires=['particle_position_x'], filtered_type='io')
        def left(pfilter, data):
            return data[(pfilter.filtered_type, 'particle_position_x')].to('code_length') < 0.5
        ds.add_particle_filter('left')

        for f in (('index', 'x'), ('index', 'dx'), ('gas', 'density')):
            ds.add_mesh_sampling_particle_field(f, ptype='io')
            ds.add_mesh_sampling_particle_field(f, ptype='left')

        data_sources = (ds.all_data(), ds.box([0]*3, [0.1]*3))

        def test_source(ptype, src):
            # Test accessing
            src[ptype, 'cell_index_x']
            src[ptype, 'cell_index_dx']
            src[ptype, 'cell_gas_density']

        for ptype in ('io', 'left'):
            for src in data_sources:
                test_source(ptype, src)

@requires_file(RAMSES)
def test_mesh_sampling_with_indexing():
    # Access with index caching
    ds = yt.load(RAMSES)
    ds.add_mesh_sampling_particle_field(('gas', 'density'), ptype='all')

    ad = ds.all_data()
    ad['all', 'cell_index']
    v1 = ad['all', 'cell_gas_density']

    # Access with no index caching
    ds = yt.load(RAMSES)
    ds.add_mesh_sampling_particle_field(('gas', 'density'), ptype='all')

    ad = ds.all_data()
    v2 = ad['all', 'cell_gas_density']

    # Check same answer is returned
    assert_allclose(v1, v2)

@requires_file(RAMSES_small)
def test_mesh_sampling_vs_field_value_at_point():
    all_ds = (
        fake_random_ds(ndims=3, particles=500),
        yt.load(RAMSES_small)
    )

    for ds in all_ds:
        ds.add_mesh_sampling_particle_field(('gas', 'density'), ptype='all')

        val = ds.r['all', 'cell_gas_density']
        ref = ds.find_field_values_at_points(('gas', 'density'), ds.r['all', 'particle_position'])

        assert_allclose(val, ref)
